"""
Test utilities for :py:mod:`betty.cache`.
"""

from collections.abc import Iterator, Sequence
from contextlib import AbstractAsyncContextManager

import pytest

from betty.cache import Cache, CacheItem


class CacheTestBase[CacheItemValueT]:
    """
    A base class for tests of :py:class:`betty.cache.Cache` implementations.
    """

    def _new_sut(
        self,
        *,
        scopes: Sequence[str] | None = None,
    ) -> AbstractAsyncContextManager[Cache[CacheItemValueT]]:
        raise NotImplementedError

    def _values(self) -> Iterator[CacheItemValueT]:
        raise NotImplementedError

    @staticmethod
    def _scopes() -> pytest.MarkDecorator:
        return pytest.mark.parametrize(
            "scopes",
            [
                (),
                ("scopey", "dopey"),
            ],
        )

    async def test_with_scope(self) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.with_scope`.
        """
        for value in self._values():
            async with self._new_sut() as sut:
                sut_with_scope_one = sut.with_scope("scopey")
                sut_with_scope_two = sut.with_scope("scopey")
                assert sut_with_scope_one is not sut
                assert sut_with_scope_two is not sut
                cache_item_id = "hello-world"
                await sut_with_scope_one.set(cache_item_id, value)
                cache_item = await sut_with_scope_two.get(cache_item_id)
                assert cache_item
                assert await cache_item.value() == value

    @_scopes()
    async def test_has__without_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.has`.
        """
        async with self._new_sut(scopes=scopes) as sut:
            assert not await sut.get("id")

    @_scopes()
    async def test_has__with_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.has`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                await sut.set("id", value)
                assert await sut.has("id")

    @_scopes()
    async def test_hasset__without_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.hasset`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                async with sut.hasset("id") as result:
                    assert result is not None
                    await result(value)
                cache_item = await sut.get("id")
                assert cache_item is not None
                assert await cache_item.value() == value

    @_scopes()
    async def test_hasset__with_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.hasset`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                await sut.set("id", value)
                async with sut.hasset("id") as result:
                    assert result is None

    @_scopes()
    async def test_get__without_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.get`.
        """
        async with self._new_sut(scopes=scopes) as sut:
            cache_item = await sut.get("id")
            assert cache_item is None

    @_scopes()
    async def test_set__and_get(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.get` and :py:meth:`betty.cache.Cache.set`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                await sut.set("id", value)
                cache_item = await sut.get("id")
                assert cache_item is not None
                assert await cache_item.value() == value

    @_scopes()
    async def test_set__and_get_with_modified(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.get` and :py:meth:`betty.cache.Cache.set`.
        """
        modified = 123456789
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                await sut.set("id", value, modified=modified)
                cache_item = await sut.get("id")
                assert cache_item is not None
                assert cache_item.modified == modified

    @_scopes()
    async def test_getset__without_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.getset`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                async with sut.getset("id") as result:
                    assert not isinstance(result, CacheItem)
                    await result(value)
                cache_item = await sut.get("id")
                assert cache_item is not None
                assert await cache_item.value() == value

    @_scopes()
    async def test_getset__with_hit(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.getset`.
        """
        for value in self._values():
            async with self._new_sut(scopes=scopes) as sut:
                await sut.set("id", value)
                async with sut.getset("id") as result:
                    assert isinstance(result, CacheItem)
                    assert await result.value() == value

    @_scopes()
    async def test_delete(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.delete`.
        """
        async with self._new_sut(scopes=scopes) as sut:
            await sut.set("id", next(self._values()))
            await sut.delete("id")
            assert await sut.get("id") is None

    @_scopes()
    async def test_clear(self, scopes: Sequence[str]) -> None:
        """
        Test implementations of :py:meth:`betty.cache.Cache.clear`.
        """
        async with self._new_sut(scopes=scopes) as sut:
            await sut.set("id", next(self._values()))
            await sut.clear()
            assert await sut.get("id") is None



"""
Provide the Render API.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, final

from betty.definition.human_facing import HumanFacingDefinition
from betty.html import plain_text_to_html
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition

if TYPE_CHECKING:
    from betty.media_type import MediaType


class Renderer(ABC, Plugin["RendererDefinition"]):
    """
    Render content in a different media type to HTML.
    """

    @property
    @abstractmethod
    def media_type(self) -> MediaType:
        """
        The media type this renderer can render from.
        """

    @abstractmethod
    async def render(self, content: str, /) -> str:
        """
        Render content.
        """


@final
@PluginTypeDefinition(
    "renderer",
    label=_("Renderer"),
    label_plural=_("Renderers"),
    label_countable=ngettext("{count} renderer", "{count} renderers"),
)
class RendererDefinition(HumanFacingDefinition, PluginDefinition[Renderer]):
    """
    .. plugin_type:: renderer.
    """


@final
class RenderDispatcher:
    """
    Dispatch content to a renderer that supports it.

    Unsupported content is rendered as plain text to make it safe for inclusion in HTML.
    """

    def __init__(self, *renderers: Renderer):
        self._renderers = {renderer.media_type: renderer for renderer in renderers}

    async def render(self, content: str, media_type: MediaType, /) -> str:
        """
        Render the content.
        """
        try:
            renderer = self._renderers[media_type]
        except KeyError:
            return plain_text_to_html(content)
        else:
            return await renderer.render(content)



"""
Provide media types.
"""

from __future__ import annotations

from betty.media_type import MediaType

HTML = MediaType("text/html", extensions=[".html"])
"""
The media type for HTML content.
"""

JINJA2 = MediaType("text/x.betty.jinja2", extensions=[".j2"])
"""
The media type for Jinja2 template files.
"""

JSON = MediaType("application/json", extensions=[".json"])
"""
The media type for JSON content.
"""

JSON_LD = MediaType("application/ld+json", extensions=[".json"])
"""
The media type for JSON-LD content.
"""

PDF = MediaType("application/pdf", extensions=[".pdf"])
"""
The media type for PDF documents.
"""

PLAIN_TEXT = MediaType("text/plain", extensions=[".txt"])
"""
The media type for plain text content.
"""

SVG = MediaType("image/svg+xml", extensions=[".svg"])
"""
The media type for SVG images.
"""

YAML = MediaType("application/yaml", extensions=[".yaml", ".yml"])
"""
The media type for YAML content.
"""



"""
Keyed collection properties.
"""

from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, override

from betty.collection.keyed import MutableKeyedCollection
from betty.functools import passthrough
from betty.property import Property

if TYPE_CHECKING:
    from ty_extensions import Intersection

    from betty.data import Data, DataDefinition
    from betty.data.aggregate.collection.keyed import KeyedCollectionDefinition
    from betty.locale.localizable import ResolvableLocalizable


class KeyedCollectionProperty[
    MutableKeyedCollectionT: MutableKeyedCollection,
    ValueSetT,
](Property[MutableKeyedCollectionT, Iterable[ValueSetT]]):
    """
    A property that contains an :py:class:`betty.collection.keyed.KeyedCollection`.
    """

    _data: KeyedCollectionDefinition[MutableKeyedCollectionT]

    def __init__(
        self,
        data: Intersection[
            DataDefinition[MutableKeyedCollectionT], KeyedCollectionDefinition
        ]
        | Data[
            Intersection[
                DataDefinition[MutableKeyedCollectionT], KeyedCollectionDefinition
            ]
        ],
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
        omit_load: bool | None = None,
        omit_dump: Callable[[MutableKeyedCollectionT], bool] | None = None,
        resolver: Callable[
            [ValueSetT | Iterable[ValueSetT]], Iterable[ValueSetT]
        ] = passthrough,
        default: Callable[[], ValueSetT | Iterable[ValueSetT]] = list,
    ):
        super().__init__(
            data,
            label=label,
            description=description,
            omit_load=omit_load,
            omit_dump=omit_dump,
            default=self._new_default,
        )
        self._values_resolver = resolver
        self._default_values = default

    def _new_default(self) -> MutableKeyedCollectionT:
        new = self._data.new()
        new.add(*self._values_resolver(self._default_values()))
        return new

    @override
    def set(
        self, instance: Any, value: Iterable[ValueSetT] | MutableKeyedCollectionT
    ) -> MutableKeyedCollectionT:
        data = self.get(instance)
        data.clear()
        data.add(*self._values_resolver(value))
        return data



"""
Provide `JSON-LD <https://json-ld.org/>`_ utilities.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import MutableSequence
from inspect import getmembers
from typing import TYPE_CHECKING, cast, final, override

from betty.classtools import Singleton
from betty.json.schema import Object, Schema
from betty.portable import PortableData, PortableMapping
from betty.string import snake_case_to_lower_camel_case

if TYPE_CHECKING:
    from betty.ancestry.link import Link
    from betty.locale.localizable import ResolvableLocalizable
    from betty.project import Project


async def dump_schema(
    project: Project,
    portable: PortableMapping,
    linked_data_dumpable: LinkedDataDumpableWithSchema[Object, PortableMapping],
    /,
) -> None:
    """
    Add the $schema item to a JSON-LD dump.
    """
    from betty.project.schema import ProjectSchema

    schema = await linked_data_dumpable.linked_data_schema(project)
    if schema.def_name:
        portable["$schema"] = await ProjectSchema.def_url(project, schema.def_name)


class LinkedDataDumpable[PortableDataT: PortableData = PortableData]:
    """
    Describe an object that can be dumped to linked data.
    """

    @abstractmethod
    async def dump_linked_data(self, project: Project, /) -> PortableDataT:
        """
        Dump this instance to `JSON-LD <https://json-ld.org/>`_.
        """


class LinkedDataDumpableWithSchema[
    SchemaT: Schema,
    PortableDataT: PortableData = PortableData,
](LinkedDataDumpable[PortableDataT]):
    """
    Describe an object that can be dumped to linked data.
    """

    @classmethod
    @abstractmethod
    async def linked_data_schema(cls, project: Project, /) -> SchemaT:
        """
        Define the `JSON Schema <https://json-schema.org/>`_ for :py:meth:`betty.json.linked_data.LinkedDataDumpable.dump_linked_data`.
        """


class JsonLdObject(Object):
    """
    A JSON Schema for an object with JSON-LD.
    """

    def __init__(
        self,
        *,
        def_name: str | None = None,
        title: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            def_name=def_name,
            title=title,
            description=description,
        )
        self._schema["allOf"] = [JsonLdSchema().embed(self)]


class LinkedDataDumpableWithSchemaJsonLdObject(
    LinkedDataDumpableWithSchema[JsonLdObject, PortableMapping], ABC
):
    """
    A :py:class:`betty.json.linked_data.LinkedDataDumpable` implementation for object/mapping data.

    This is helpful when working with diamond class hierarchies where parent classes that may not be the root class want
    to make changes to the linked data, and expect an :py:class`betty.json.schema.Object` schema and a
    :py:type:`betty.portable.PortableMapping` dump.
    """

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = JsonLdObject()
        for attr_name, class_attr_value in getmembers(cls):
            if isinstance(class_attr_value, LinkedDataDumper):
                linked_data_dumpable = class_attr_value
                schema.add_property(
                    snake_case_to_lower_camel_case(attr_name),
                    await linked_data_dumpable.linked_data_schema_for(project),
                    True,
                )
        return schema

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable: PortableMapping = {}

        await dump_schema(project, portable, self)

        for attr_name, class_attr_value in getmembers(type(self)):
            if isinstance(class_attr_value, LinkedDataDumper):
                portable[
                    snake_case_to_lower_camel_case(attr_name)
                ] = await class_attr_value.dump_linked_data_for(project, self)

        return portable


class LinkedDataDumper[
    T,
    SchemaT: Schema = Schema,
    PortableDataT: PortableData = PortableData,
](ABC):
    """
    Provide linked data for instances of a target type.
    """

    @abstractmethod
    async def linked_data_schema_for(self, project: Project, /) -> SchemaT:
        """
        Define the `JSON Schema <https://json-schema.org/>`_ for :py:meth:`betty.json.linked_data.LinkedDataDumper.dump_linked_data_for`.
        """

    @abstractmethod
    async def dump_linked_data_for(
        self, project: Project, target: T, /
    ) -> PortableDataT:
        """
        Dump the given target to `JSON-LD <https://json-ld.org/>`_.
        """


def dump_context(portable: PortableMapping, **context_definitions: str) -> None:
    """
    Add one or more contexts to a dump.
    """
    portable_context = cast(PortableMapping, portable.setdefault("@context", {}))
    for key, context_definition in context_definitions.items():
        portable_context[key] = context_definition


async def dump_link(portable: PortableMapping, project: Project, *links: Link) -> None:
    """
    Add one or more links to a dump.
    """
    portable_link = cast(
        MutableSequence[PortableMapping],
        portable.setdefault("links", []),
    )
    for link in links:
        portable_link.append(await link.dump_linked_data(project))


@final
class JsonLdSchema(Singleton, Schema):
    """
    A `JSON-LD <https://json-ld.org/>`_ JSON Schema reference.
    """

    _SCHEMA: PortableMapping = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "additionalProperties": True,
        "allOf": [
            {"$ref": "#/$defs/context"},
            {"$ref": "#/$defs/graph"},
            {"$ref": "#/$defs/common"},
        ],
        "$defs": {
            "context": {
                "additionalProperties": True,
                "properties": {
                    "@context": {
                        "description": "Used to define the short-hand names that are used throughout a JSON-LD document.",
                        "type": ["object", "string", "array", "null"],
                    }
                },
            },
            "graph": {
                "additionalProperties": True,
                "properties": {
                    "@graph": {
                        "description": "Used to express a graph.",
                        "anyOf": [
                            {"type": "array", "items": {"$ref": "#/$defs/common"}},
                            {"$ref": "#/$defs/common", "type": "object"},
                        ],
                    }
                },
            },
            "common": {
                "additionalProperties": {"anyOf": [{"$ref": "#/$defs/common"}]},
                "properties": {
                    "@id": {
                        "description": "Used to uniquely identify things that are being described in the document with IRIs or blank node identifiers.",
                        "type": "string",
                        "format": "uri",
                    },
                    "@value": {
                        "description": "Used to specify the data that is associated with a particular property in the graph.",
                        "type": ["string", "boolean", "number", "null"],
                    },
                    "@language": {
                        "description": "Used to specify the language for a particular string value or the default language of a JSON-LD document.",
                        "type": ["string", "null"],
                    },
                    "@type": {
                        "description": "Used to set the data type of a node or typed value.",
                        "type": ["string", "null", "array"],
                    },
                    "@container": {
                        "description": "Used to set the default container type for a term.",
                        "type": ["string", "null"],
                        "enum": ["@language", "@list", "@index", "@set"],
                    },
                    "@list": {"description": "Used to express an ordered set of data."},
                    "@set": {
                        "description": "Used to express an unordered set of data and to ensure that values are always represented as arrays."
                    },
                    "@reverse": {
                        "description": "Used to express reverse properties.",
                        "type": ["string", "object", "null"],
                        "additionalProperties": {"anyOf": [{"$ref": "#/$defs/common"}]},
                    },
                    "@base": {
                        "description": "Used to set the base IRI against which relative IRIs are resolved",
                        "type": ["string", "null"],
                        "format": "uri",
                    },
                    "@vocab": {
                        "description": "Used to expand properties and values in @type with a common prefix IRI",
                        "type": ["string", "null"],
                        "format": "uri",
                    },
                },
            },
        },
        "title": "Schema for JSON-LD",
        "type": ["object", "array"],
    }

    def __init__(self):
        super().__init__(def_name="jsonLd", title="JSON-LD")
        self._schema = self._SCHEMA



"""
Test utilities for :py:mod:`betty.model.collections`.
"""

from collections.abc import Sequence
from typing import Any

import pytest

from betty.model import Entity
from betty.model.collections import EntityCollection


class EntityCollectionTestBase[EntityT: Entity]:
    """
    A base class for testing :py:class:`betty.model.collections.EntityCollection` implementations.
    """

    @pytest.fixture
    def sut(self) -> EntityCollection[EntityT]:
        """
        Provide the system(s) under test.
        """
        raise NotImplementedError

    @pytest.fixture
    async def sut_entities(self) -> Sequence[EntityT]:
        """
        Produce entities to test the collections with.

        This MUST return at least 3 entities.
        """
        raise NotImplementedError

    async def test_entity_collection_test_base_sut_entities(
        self, sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.test_utils.model.collections.EntityCollectionTestBase.sut_entities` implementations.
        """
        assert len(sut_entities) >= 3

    async def test_add(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.add` implementations.
        """
        sut.add(*sut_entities)
        assert list(sut) == list(sut_entities)

    async def test_add_with_duplicate_entities(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.add` implementations.
        """
        sut.add(sut_entities[0], sut_entities[1], sut_entities[0], sut_entities[2])
        assert list(sut) == list(sut_entities[0:3])

    async def test_remove(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.remove` implementations.
        """
        sut.add(*sut_entities)
        first = sut_entities[0]
        sut.add(*sut_entities)
        sut.remove(first)
        assert list(sut) == list(sut_entities[1:])
        sut.remove(*sut_entities)
        assert list(sut) == []

    async def test___delitem____by_entity(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__delitem__` implementations.
        """
        sut.add(*sut_entities)
        first = sut_entities[0]
        sut.add(*sut_entities)
        del sut[first]
        assert list(sut) == list(sut_entities[1:])

    async def test___contains____by_entity(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__contains__` implementations.
        """
        sut.add(sut_entities[0])
        assert sut_entities[0] in sut
        assert sut_entities[1] not in sut

    @pytest.mark.parametrize(
        "value",
        [
            True,
            False,
            [],
        ],
    )
    async def test___contains____by_unsupported_type(
        self, value: Any, sut: EntityCollection[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__contains__` implementations.
        """
        assert value not in sut

    async def test___len__(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__len__` implementations.
        """
        assert len(sut) == 0
        sut.add(*sut_entities)
        assert len(sut) == len(sut_entities)

    async def test___iter__(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__iter__` implementations.
        """
        assert list(iter(sut)) == []
        sut.add(*sut_entities)
        assert list(iter(sut)) == list(sut_entities)

    async def test_clear(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.__iter__` implementations.
        """
        sut.add(*sut_entities)
        sut.clear()
        assert list(sut) == []

    async def test_replace(
        self, sut: EntityCollection[EntityT], sut_entities: Sequence[EntityT]
    ) -> None:
        """
        Tests :py:meth:`betty.model.collections.EntityCollection.replace` implementations.
        """
        first = sut_entities[0]
        others = sut_entities[1:]
        sut.add(first)
        sut.replace(*others)
        assert list(sut) == list(others)



"""
Data types for plugins.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, final

from betty.collection.sequence import (
    MutableResolvedSequence,
)
from betty.collection.sequence.adapter import MutableResolvedSequenceAdapter
from betty.data import Data
from betty.data.aggregate.collection.sequence import SequenceDefinition
from betty.data.aggregate.record.object import ObjectDefinition
from betty.locale.localizable.gettext import _
from betty.locale.localizable.property import (
    CountableLocalizableProperty,
    LocalizableProperty,
)
from betty.machine_name import MachineNameProperty, ResolvableMachineName
from betty.plugin import PluginDefinition
from betty.property import Optional

if TYPE_CHECKING:
    from betty.locale.localizable import (
        ResolvableCountableLocalizable,
        ResolvableLocalizable,
    )
    from betty.plugin.factory import PluginManufacturer


class PluginDefinitionConfiguration[
    PluginDefinitionT: PluginDefinition = PluginDefinition
](Data[ObjectDefinition["PluginDefinitionConfiguration"]], ABC):
    """
    Configure a :py:class:`betty.plugin.PluginDefinition`.

    .. data:: betty.plugin.config:PluginDefinitionConfiguration
    """

    id = MachineNameProperty(label=_("Plugin ID"))
    """
    The plugin ID.
    """

    def __init__(
        self,
        *,
        id: ResolvableMachineName,  # noqa: A002
    ):
        super().__init__()
        self.id = id

    @abstractmethod
    def new_plugin(self) -> PluginDefinitionT:
        """
        Create a new plugin from this configuration.
        """


class HumanFacingPluginDefinitionConfiguration[
    PluginDefinitionT: PluginDefinition = PluginDefinition
](PluginDefinitionConfiguration[PluginDefinitionT]):
    """
    Configure a :py:class:`betty.definition.human_facing.HumanFacingDefinition`.

    .. data:: betty.plugin.config:HumanFacingPluginDefinitionConfiguration
    """

    label = LocalizableProperty(label=_("Label"))
    description = Optional(LocalizableProperty(label=_("Description")))

    def __init__(
        self,
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.label = label
        self.description = description


class CountableHumanFacingPluginDefinitionConfiguration[
    PluginDefinitionT: PluginDefinition = PluginDefinition
](HumanFacingPluginDefinitionConfiguration[PluginDefinitionT]):
    """
    Configure a :py:class:`betty.definition.human_facing.CountableHumanFacingDefinition`.

    .. data:: betty.plugin.config:CountableHumanFacingPluginDefinitionConfiguration
    """

    label_plural = LocalizableProperty(label=_("Label (plural)"))
    label_countable = CountableLocalizableProperty(label=_("Label (countable)"))

    def __init__(
        self,
        *,
        label_plural: ResolvableLocalizable,
        label_countable: ResolvableCountableLocalizable,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.label_plural = label_plural
        self.label_countable = label_countable


@final
class PluginManufacturerSequenceDefinition(SequenceDefinition):
    """
    Define a sequence of plugin instance configurations.
    """

    def __init__(
        self,
        manufacturer: type[PluginManufacturer],
        *,
        label: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            cls=MutableResolvedSequence,
            factory=lambda: MutableResolvedSequenceAdapter(
                [], value_resolver=manufacturer.resolve
            ),
            value=manufacturer,
            label=manufacturer.type().type().label_plural if label is None else label,
        )



"""
Provide a subprocess API.
"""

import os
import subprocess
from asyncio import create_subprocess_exec, create_subprocess_shell
from asyncio.subprocess import Process
from collections.abc import Sequence
from pathlib import Path
from subprocess import PIPE

from betty.locale.localizable.gettext import _
from betty.user import User, Verbosity


class SubprocessError(Exception):
    """
    Raised when a subprocess failed.
    """


class CalledSubprocessError(subprocess.CalledProcessError, SubprocessError):
    """
    Raised when a subprocess was successfully invoked, but subsequently failed during its own execution.
    """


class FileNotFound(FileNotFoundError, SubprocessError):
    """
    Raised when a command could not be found.
    """


async def run_process(
    runnee: Sequence[str], cwd: Path | None = None, shell: bool = False, *, user: User
) -> Process:
    """
    Run a command in a subprocess.

    :raise betty.subprocess.SubprocessError:
    """
    command = " ".join(runnee)
    await user.message_debug(
        _("Running subprocess `{command}`...").format(command=command)
    )

    try:
        if shell:
            process = await create_subprocess_shell(
                " ".join(runnee), cwd=cwd, stderr=PIPE, stdout=PIPE
            )
        else:
            process = await create_subprocess_exec(
                *runnee, cwd=cwd, stderr=PIPE, stdout=PIPE
            )
        stdout, stderr = await process.communicate()
    except FileNotFoundError as error:
        await user.message_debug(str(error))
        raise FileNotFound(str(error)) from None

    if process.returncode != 0 or user.verbosity is Verbosity.MOST_VERBOSE:
        stdout_str = "\n".join(stdout.decode().split(os.linesep))
        stderr_str = "\n".join(stderr.decode().split(os.linesep))

        if stdout_str:
            await user.message_debug(
                _("Subprocess `{command}` stdout:\n{stdout}").format(
                    command=command, stdout=stdout_str
                )
            )
        if stderr_str:
            await user.message_debug(
                _("Subprocess `{command}` stderr:\n{stderr}").format(
                    command=command, stderr=stderr_str
                )
            )

        if process.returncode != 0:
            assert process.returncode is not None
            raise CalledSubprocessError(
                process.returncode,
                " ".join(runnee),
                stdout_str,
                stderr_str,
            )

    return process



"""
Provide copyright notices.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, final, override

from betty.definition.human_facing import HumanFacingDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.factory import PluginManufacturer

if TYPE_CHECKING:
    import builtins

    from betty.locale.localizable import Localizable


class CopyrightNotice(Plugin["CopyrightNoticeDefinition"]):
    """
    A copyright notice.

    To test your own subclasses, use :py:class:`betty.test_utils.copyright_notice.CopyrightNoticeTestBase`.
    """

    @property
    @abstractmethod
    def summary(self) -> Localizable:
        """
        The copyright summary.
        """

    @property
    @abstractmethod
    def text(self) -> Localizable:
        """
        The full copyright text.
        """

    @property
    def url(self) -> Localizable | None:
        """
        The URL to an external human-readable resource with more information about this copyright.
        """
        return None


@final
@PluginTypeDefinition(
    "copyright-notice",
    label=_("Copyright notice"),
    label_plural=_("Copyright notices"),
    label_countable=ngettext("{count} copyright notice", "{count} copyright notices"),
)
class CopyrightNoticeDefinition(
    HumanFacingDefinition, PluginDefinition[CopyrightNotice]
):
    """
    .. plugin_type:: copyright-notice.
    """


@final
class CopyrightNoticeManufacturer(
    PluginManufacturer[CopyrightNoticeDefinition, CopyrightNotice]
):
    """
    The copyright notice manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[CopyrightNoticeDefinition]:
        return CopyrightNoticeDefinition



"""
Color management.
"""

from __future__ import annotations

import re
from typing import final

from betty.assertion import assert_str
from betty.data import DataDefinition, Sample
from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.portable import CallbackPorter

_HEX_PATTERN = re.compile(r"^#[a-zA-Z0-9]{6}$")


def _assert_hex(color: str) -> str:
    if not _HEX_PATTERN.match(color):
        raise HumanFacingException(
            _('"{color}" is not a valid hexadecimal color, such as #ffc0cb.').format(
                color=color,
            )
        )
    return color


@final
class ColorDefinition(DataDefinition):
    """
    Define a color.
    """

    def __init__(self):
        super().__init__(
            cls=str,
            label=_("Color"),
            description=_("A hexadecimal color, such as #ff0000"),
            samples=[lambda: Sample("#ff0000", label="Default")],
            porter=CallbackPorter(assert_str() | _assert_hex, str),
        )



"""
Localizers provide a wide range of localization utilities through a single entry point.
"""

from __future__ import annotations

import gettext as gettext_api
from typing import TYPE_CHECKING, final

from betty.locale import (
    DEFAULT_LOCALE,
    HasLocale,
    HasLocaleStr,
    ResolvableLocale,
    resolve_locale,
)
from betty.typing import threadsafe

if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from babel import Locale
    from ty_extensions import Intersection

    from betty.locale.localizable import ResolvableLocalizable
    from betty.locale.translation import TranslationRepository


@final
class Localizer:
    """
    Localize a variety of data into a specific locale.
    """

    def __init__(
        self, locale: ResolvableLocale, translations: gettext_api.NullTranslations, /
    ):
        self._locale = resolve_locale(locale)
        self._translations = translations

    @property
    def locale(self) -> Locale:
        """
        The locale.
        """
        return self._locale

    def _(self, message: str, /) -> Intersection[HasLocale, str]:
        """
        Like :py:meth:`gettext.gettext`.

        Arguments are identical to those of :py:meth:`gettext.gettext`.
        """
        return HasLocaleStr(self._translations.gettext(message), locale=self._locale)

    def gettext(self, message: str, /) -> Intersection[HasLocale, str]:
        """
        Like :py:meth:`gettext.gettext`.

        Arguments are identical to those of :py:meth:`gettext.gettext`.
        """
        return HasLocaleStr(self._translations.gettext(message), locale=self._locale)

    def ngettext(
        self, message_singular: str, message_plural: str, n: int, /
    ) -> Intersection[HasLocale, str]:
        """
        Like :py:meth:`gettext.ngettext`.

        Arguments are identical to those of :py:meth:`gettext.ngettext`.
        """
        return HasLocaleStr(
            self._translations.ngettext(message_singular, message_plural, n),
            locale=self._locale,
        )

    def pgettext(self, context: str, message: str, /) -> Intersection[HasLocale, str]:
        """
        Like :py:meth:`gettext.pgettext`.

        Arguments are identical to those of :py:meth:`gettext.pgettext`.
        """
        return HasLocaleStr(
            self._translations.pgettext(context, message), locale=self._locale
        )

    def npgettext(
        self, context: str, message_singular: str, message_plural: str, n: int, /
    ) -> Intersection[HasLocale, str]:
        """
        Like :py:meth:`gettext.npgettext`.

        Arguments are identical to those of :py:meth:`gettext.npgettext`.
        """
        return HasLocaleStr(
            self._translations.npgettext(context, message_singular, message_plural, n),
            locale=self._locale,
        )


DEFAULT_LOCALIZER = Localizer(DEFAULT_LOCALE, gettext_api.NullTranslations())


@final
@threadsafe
class LocalizerRepository:
    """
    Exposes the available localizers.
    """

    def __init__(self, translations: TranslationRepository, /):
        self._translations = translations
        self._localizers: MutableMapping[Locale, Localizer] = {}

    def get(self, locale: ResolvableLocale, /) -> Localizer:
        """
        Get the localizer for the given locale.
        """
        locale = resolve_locale(locale)
        try:
            return self._localizers[locale]
        except KeyError:
            from betty.locale.translation import UntranslatedLocale

            try:
                translations = self._translations.get(locale)
            except UntranslatedLocale:
                translations = gettext_api.NullTranslations()
            self._localizers[locale] = Localizer(locale, translations)
            return self._localizers[locale]


def resolve_localized(
    localizable: ResolvableLocalizable, *, localizer: Localizer
) -> Intersection[HasLocale, str]:
    """
    Ensure that a localizable-like value is or is made to be localized.
    """
    from betty.locale.localizable import Localizable

    if isinstance(localizable, str):
        return HasLocaleStr(localizable)
    if not isinstance(localizable, Localizable):
        from betty.locale.localizable.static import StaticTranslations

        localizable = StaticTranslations(localizable)
    return localizable.localize(localizer)



"""
Common theme functionality.

The contents of this file should eventually be stabilized and moved to more specific modules.
"""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, cast

from betty.ancestry.event import Event
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.date import Date, ResolvableDate
from betty.event_type.event_types import Birth, Death
from betty.functools import unique
from betty.model import persistent_id
from betty.privacy import is_public
from betty.role.roles import Subject

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from betty.ancestry.file_reference import FileReference
    from betty.ancestry.has_file_references import HasFileReferences
    from betty.ancestry.presence import Presence
    from betty.jinja import Filters
    from betty.project import Project


def _is_person_timeline_presence(presence: Presence) -> bool:
    if presence.private:
        return False
    if not presence.event.date:
        return False
    return presence.event.date.comparable


def person_timeline_events(person: Person, lifetime_threshold: int) -> Iterable[Event]:
    """
    Gather all events for a person's timeline.
    """
    yield from unique(_person_timeline_events(person, lifetime_threshold))


def person_descendant_families(
    person: Person,
) -> Iterable[tuple[Sequence[Person], Sequence[Person]]]:
    """
    Gather a person's families they are a parent in.
    """
    parents = {}
    children = defaultdict(set)
    for child in person.children:
        family = tuple(sorted(parent.id for parent in child.parents))
        if family not in parents:
            parents[family] = tuple(child.parents)
        children[family].add(child)
    yield from zip(parents.values(), children.values(), strict=True)


def associated_file_references(
    has_file_references: HasFileReferences,
) -> Iterable[FileReference]:
    """
    Get the associated file references for an entity that has file references.
    """
    yield from unique(
        _associated_file_references(has_file_references),
        key=lambda file_reference: file_reference.file,
    )


def _associated_file_references(
    has_file_references: HasFileReferences,
) -> Iterable[FileReference]:
    yield from has_file_references.file_references

    if isinstance(has_file_references, Event):
        for citation in has_file_references.citations:
            yield from _associated_file_references(citation)

    if isinstance(has_file_references, Person):
        for name in has_file_references.names:
            for citation in name.citations:
                yield from _associated_file_references(citation)
        for presence in has_file_references.presences:
            yield from _associated_file_references(presence.event)

    if isinstance(has_file_references, Place):
        for event in has_file_references.events:
            yield from _associated_file_references(event)


def _person_timeline_events(person: Person, lifetime_threshold: int) -> Iterable[Event]:
    # Collect all associated events for a person.
    # Start with the person's own events for which their presence is public.
    for presence in person.presences:
        if _is_person_timeline_presence(presence):
            assert presence.event is not None
            yield presence.event
        continue

    # If the person has start- or end-of-life events, we use those to constrain associated people's events.
    start_dates = []
    end_dates = []
    for presence in person.presences:
        if not _is_person_timeline_presence(presence):
            continue
        assert presence.event is not None
        assert presence.event.date is not None
        if presence.role.plugin().id != Subject.plugin().id:
            continue
        if (
            presence.event.event_type.plugin().id == Birth.plugin().id
            or presence.event.event_type.plugin().indicates == Birth.plugin().id
        ):
            start_dates.append(presence.event.date)
        if (
            presence.event.event_type.plugin().id == Death.plugin().id
            or presence.event.event_type.plugin().indicates == Death.plugin().id
        ):
            end_dates.append(presence.event.date)
    start_date = sorted(start_dates)[0] if start_dates else None
    end_date = sorted(end_dates)[0] if end_dates else None

    # If an end-of-life event exists, but no start-of-life event, create a start-of-life date based on the end date,
    # minus the lifetime threshold.
    if start_date is None and end_date is not None:
        if isinstance(end_date, Date):
            start_date_reference = end_date
        else:
            if end_date.end is not None and end_date.end.comparable:
                start_date_reference = end_date.end
            else:
                assert end_date.start is not None
                start_date_reference = end_date.start
        assert start_date_reference.year is not None
        start_date = Date(
            start_date_reference.year - lifetime_threshold,
            start_date_reference.month,
            start_date_reference.day,
            fuzzy=start_date_reference.fuzzy,
        )

    # If a start-of-life event exists, but no end-of-life event, create an end-of-life date based on the start date,
    # plus the lifetime threshold.
    if end_date is None and start_date is not None:
        if isinstance(start_date, Date):
            end_date_reference = start_date
        else:
            if start_date.start and start_date.start.comparable:
                end_date_reference = start_date.start
            else:
                assert start_date.end is not None
                end_date_reference = start_date.end
        assert end_date_reference.year is not None
        end_date = Date(
            end_date_reference.year + lifetime_threshold,
            end_date_reference.month,
            end_date_reference.day,
            fuzzy=end_date_reference.fuzzy,
        )

    if start_date is None or end_date is None:
        reference_dates = sorted(
            cast(ResolvableDate, presence.event.date)
            for presence in person.presences
            if _is_person_timeline_presence(presence)
        )
        if reference_dates:
            if not start_date:
                start_date = reference_dates[0]
            if not end_date:
                end_date = reference_dates[-1]

    if start_date is not None and end_date is not None:
        associated_people = filter(
            is_public,
            (
                # All ancestors.
                *person.ancestors,
                # All descendants.
                *person.descendants,
                # All siblings.
                *person.siblings,
            ),
        )
        for associated_person in associated_people:
            # For associated events, we are only interested in people's start- or end-of-life events.
            for associated_presence in associated_person.presences:
                if (
                    associated_presence.event.event_type.plugin().id
                    != Birth.plugin().id
                    and associated_presence.event.event_type.plugin().id
                    != Birth.plugin().id
                    and associated_presence.event.event_type.plugin().id
                    != Death.plugin().id
                    and associated_presence.event.event_type.plugin().indicates
                    != Death.plugin().id
                ):
                    continue
                if not persistent_id(associated_presence.event):
                    continue
                if not isinstance(associated_presence.role, Subject):
                    continue
                if not _is_person_timeline_presence(associated_presence):
                    continue
                if not associated_presence.event.date:
                    continue
                if associated_presence.event.date < start_date:
                    continue
                if associated_presence.event.date > end_date:
                    continue
                yield associated_presence.event


def jinja_filters(project: Project) -> Filters:
    return {
        "person_timeline_events": lambda person: person_timeline_events(
            person, project.configuration.lifetime_threshold
        ),
        "person_descendant_families": person_descendant_families,
        "associated_file_references": associated_file_references,
    }


def place_timeline_events(place: Place, /) -> Iterable[Event]:
    """
    Gather all events for a person's timeline.
    """
    yield from unique(_place_timeline_events(place))


def _place_timeline_events(place: Place, /) -> Iterable[Event]:
    yield from place.events
    for enclosee in place.enclosees:
        yield from enclosee.enclosee.events



from __future__ import annotations  # noqa: D100

from pathlib import Path
from typing import TYPE_CHECKING, Self, final, override

from aiofiles.os import makedirs
from aiofiles.tempfile import TemporaryDirectory

from betty import dirs
from betty.about import IS_DEVELOPMENT
from betty.app import App
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.extension.demo import generate_with_cleanup
from betty.extension.demo.project import create_project
from betty.job import Context
from betty.service.factory import Manufacturable
from betty.service.requirement import UnmetRequirement
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse
    from collections.abc import Iterable

    from yappi import YFuncStats

    from betty.plugin.discovery import ResolvableDiscovery
    from betty.user import User


async def _target(user: User) -> None:
    async with (
        App.new_isolated() as app,
        app,
        TemporaryDirectory() as project_directory_path_str,
    ):
        project = await create_project(app, Path(project_directory_path_str))
        async with project, user.message_progress("Generating site...") as progress:
            await generate_with_cleanup(project, context=Context(progress=progress))


def _print(stats: YFuncStats, sort_column: str, sort_direction: str) -> None:
    stats.sort(sort_column, sort_direction)
    stats.print_all(
        columns={
            0: ("tsub", 10),
            1: ("ttot", 10),
            2: ("tavg", 10),
            3: ("ncall", 10),
            4: ("name", 99),
        }
    )


@final
@CommandDefinition(
    "dev-profile-demo", label="Profile the generation of the demonstration site"
)
class DevProfileDemo(Manufacturable, Command):
    """
    .. plugin:: command:dev-profile-demo.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        parser.set_defaults(
            clock_type="CPU",
            force=False,
            sort_column="tsub",
            sort_direction="asc",
        )
        clock_type_group = parser.add_mutually_exclusive_group()
        clock_type_group.add_argument(
            "--cpu",
            dest="clock_type",
            action="store_const",
            const="CPU",
            help="Use the cpu clock",
        )
        clock_type_group.add_argument(
            "--wall",
            dest="clock_type",
            action="store_const",
            const="WALL",
            help="Use the wall clock",
        )
        sort_column_group = parser.add_mutually_exclusive_group()
        sort_column_group.add_argument(
            "--tsub",
            dest="sort_column",
            action="store_const",
            const="tsub",
            help="Sort on the `tsub` column",
        )
        sort_column_group.add_argument(
            "--ttot",
            dest="sort_column",
            action="store_const",
            const="ttot",
            help="Sort on the `ttot` column",
        )
        sort_column_group.add_argument(
            "--tavg",
            dest="sort_column",
            action="store_const",
            const="tavg",
            help="Sort on the `tavg` column",
        )
        sort_column_group.add_argument(
            "--ncall",
            dest="sort_column",
            action="store_const",
            const="ncall",
            help="Sort on the `ncall` column",
        )
        sort_column_group.add_argument(
            "--name",
            dest="sort_column",
            action="store_const",
            const="name",
            help="Sort on the `name` column",
        )
        sort_direction_group = parser.add_mutually_exclusive_group()
        sort_direction_group.add_argument(
            "--asc",
            dest="sort_direction",
            action="store_const",
            const="asc",
            help="Sort results in ascending order",
        )
        sort_direction_group.add_argument(
            "--desc",
            dest="sort_direction",
            action="store_const",
            const="desc",
            help="Sort results in descending order",
        )
        parser.add_argument(
            "--force",
            dest="force",
            action="store_true",
            help="Ignore the cache and create new profiling stats",
            default=False,
        )
        return self._command_function

    async def _command_function(
        self, *, clock_type: str, force: bool, sort_column: str, sort_direction: str
    ) -> None:
        import yappi

        stats_file_path = (
            dirs.DEV_OUTPUT_DIRECTORY_PATH / f"{self.plugin().id}-{clock_type}.ystats"
        )
        if not force and stats_file_path.exists():
            stats = yappi.get_func_stats()
            stats.add([stats_file_path])
            _print(stats, sort_column, sort_direction)
            await self._app.user.message_information(
                f"Showing existing stats from {stats_file_path}"
            )
        else:
            await makedirs(stats_file_path.parent, exist_ok=True)
            yappi.set_clock_type(clock_type)  # Use set_clock_type("wall") for wall time
            yappi.start()
            await _target(self._app.user)
            yappi.stop()
            stats = yappi.get_func_stats()
            stats.save(stats_file_path)
            _print(stats, sort_column, sort_direction)
            await self._app.user.message_information(
                f"Showing newly generated stats from {stats_file_path}"
            )


def _discover(_) -> Iterable[ResolvableDiscovery[CommandDefinition]]:
    if not IS_DEVELOPMENT:
        raise UnmetRequirement("This is only available when developing Betty")
    yield DevProfileDemo



"""
Complex/markup localizables.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from textwrap import indent
from typing import TYPE_CHECKING, ClassVar, final, override

from betty.locale import HasLocale, HasLocaleStr
from betty.locale.localizable import (
    Localizable,
    ResolvableLocalizable,
    resolve_localizable,
)
from betty.locale.localizable.gettext import _

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ty_extensions import Intersection

    from betty.locale.localize import Localizer


class LocalizableSequence(ABC):
    """
    A sequence of localizables.
    """

    @property
    @abstractmethod
    def localizables(self) -> Sequence[Localizable]:
        """
        The localizables.
        """


class _LocalizableSequence(LocalizableSequence):
    def __init__(self, *localizables: ResolvableLocalizable):
        self._localizables = tuple(map(resolve_localizable, localizables))

    @override
    @property
    def localizables(self) -> Sequence[Localizable]:
        return self._localizables


class _Join(_LocalizableSequence, Localizable):
    _SEPARATOR: ClassVar[str]

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        return HasLocaleStr(
            self._SEPARATOR.join(
                localized
                for part in self.localizables
                if (localized := part.localize(localizer))
            ),
            locale=localizer.locale,
        )


@final
class Chain(_Join):
    """
    Chain multiple localizables together, back to back.
    """

    _SEPARATOR = ""


@final
class Paragraph(_Join):
    """
    Represent multiple localizables as a single paragraph of text.
    """

    _SEPARATOR = " "


@final
class Lines(_Join):
    """
    Represent multiple localizables as multiple lines of text.
    """

    _SEPARATOR = "\n"


@final
class Paragraphs(_Join):
    """
    Represent multiple localizables as multiple paragraphs of text.
    """

    _SEPARATOR = "\n\n"


class _List(_LocalizableSequence, Localizable):
    _TEMPLATE_LEFT_TO_RIGHT = "{prefix} {localized}"
    _TEMPLATE_RIGHT_TO_LEFT = "{localized} {prefix}"

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        if not self.localizables:
            return HasLocaleStr("")
        localizeds = []
        prefixes = []
        prefix_lengths = []
        if localizer.locale.character_order == "right-to-left":
            template = self._TEMPLATE_RIGHT_TO_LEFT
        else:
            template = self._TEMPLATE_LEFT_TO_RIGHT
        for index, localizable in enumerate(self._localizables):
            localizeds.append(localizable.localize(localizer))
            prefix = self._get_prefix(localizer, index)
            prefixes.append(prefix)
            prefix_lengths.append(len(prefix))
        max_prefix_length = max(prefix_lengths) + 1
        return HasLocaleStr(
            "\n".join(
                template.format(
                    localized=indent(localized, " " * max_prefix_length)[
                        len(prefixes[index]) + 1 :
                    ],
                    prefix=self._get_prefix(localizer, index),
                )
                for index, localized in enumerate(localizeds)
            )
        )

    @abstractmethod
    def _get_prefix(self, localizer: Localizer, index: int, /) -> str:
        pass


@final
class OrderedList(_List):
    """
    Represent multiple localizables in an ordered list.
    """

    _PREFIX_TEMPLATE_LEFT_TO_RIGHT = "{index}."
    _PREFIX_TEMPLATE_RIGHT_TO_LEFT = ".{index}"

    @override
    def _get_prefix(self, localizer: Localizer, index: int, /) -> str:
        if localizer.locale.character_order == "right-to-left":
            template = self._PREFIX_TEMPLATE_RIGHT_TO_LEFT
        else:
            template = self._PREFIX_TEMPLATE_LEFT_TO_RIGHT
        return template.format(index=index + 1)


@final
class UnorderedList(_List):
    """
    Represent multiple localizables in an unordered list.
    """

    @override
    def _get_prefix(self, localizer: Localizer, index: int, /) -> str:
        return "-"


class _Enumeration(_LocalizableSequence, Localizable):
    _LOCALIZABLE: ClassVar[Localizable]

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        if len(self.localizables) == 0:
            return HasLocaleStr("")
        if len(self.localizables) == 1:
            return self.localizables[0].localize(localizer)
        return self._LOCALIZABLE.format(
            most=", ".join(
                part.localize(localizer) for part in self.localizables[0:-1]
            ),
            last=self.localizables[-1],
        ).localize(localizer)


@final
class AnyEnumeration(_Enumeration):
    """
    An enumeration where any of the localizables may be applicable.
    """

    _LOCALIZABLE = _("{most}, or {last}")


@final
class AllEnumeration(_Enumeration):
    """
    An enumeration where all of the localizables are applicable.
    """

    _LOCALIZABLE = _("{most}, and {last}")


def do_you_mean(*available_options: str) -> Localizable:
    """
    Produce a message listing available options.
    """
    match len(available_options):
        case 0:
            return _("There are no available options.")
        case 1:
            return _("Do you mean {available_option}?").format(
                available_option=available_options[0]
            )
        case _:
            return _("Do you mean one of {available_options}?").format(
                available_options=AnyEnumeration(*sorted(map(str, available_options)))
            )



"""
Provide licenses.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, final, override

from betty.definition.human_facing import HumanFacingDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.factory import PluginManufacturer

if TYPE_CHECKING:
    import builtins

    from betty.locale.localizable import Localizable


class License(Plugin["LicenseDefinition"]):
    """
    A license.

    To test your own subclasses, use :py:class:`betty.test_utils.license.LicenseTestBase`.
    """

    @property
    @abstractmethod
    def summary(self) -> Localizable:
        """
        The license summary.
        """

    @property
    @abstractmethod
    def text(self) -> Localizable:
        """
        The full license text.
        """

    @property
    def url(self) -> Localizable | None:
        """
        The URL to an external human-readable resource with more information about this license.
        """
        return None


@final
@PluginTypeDefinition(
    "license",
    label=_("License"),
    label_plural=_("Licenses"),
    label_countable=ngettext("{count} license", "{count} licenses"),
)
class LicenseDefinition(HumanFacingDefinition, PluginDefinition[License]):
    """
    .. plugin_type:: license.
    """


@final
class LicenseManufacturer(PluginManufacturer[LicenseDefinition, License]):
    """
    The license manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[LicenseDefinition]:
        return LicenseDefinition



"""
Keyed collection proxies that raise custom errors.
"""

from collections.abc import Callable
from typing import final, override

from betty.collection.keyed import KeyedCollection, MutableKeyedCollection
from betty.collection.keyed.proxy import (
    KeyedCollectionProxy,
    MutableKeyedCollectionProxy,
)


class _ErroringKeyedCollection[
    KeyT,
    ResolvableKeyT,
    ValueT,
    KeyedCollectionT: KeyedCollection,
](KeyedCollection[KeyT, ResolvableKeyT, ValueT]):
    def __init__(
        self,
        upstream: KeyedCollection[KeyT, ResolvableKeyT, ValueT],
        key_error: Callable[[KeyError, KeyT | ResolvableKeyT], KeyError],
        /,
    ):
        super().__init__(upstream)
        self._key_error = key_error


@final
class ErroringKeyedCollection[KeyT, ResolvableKeyT, ValueT](
    _ErroringKeyedCollection[
        KeyT, ResolvableKeyT, ValueT, KeyedCollection[KeyT, ResolvableKeyT, ValueT]
    ],
    KeyedCollectionProxy[KeyT, ResolvableKeyT, ValueT],
):
    """
    A keyed collection proxy that raises custom errors.
    """

    @override
    def __getitem__(self, key: KeyT | ResolvableKeyT) -> ValueT:
        try:
            return self._upstream[key]
        except KeyError as error:
            raise self._key_error(error, key) from error


@final
class MutableErroringKeyedCollection[KeyT, ResolvableKeyT, ValueT, ResolvableValueT](
    _ErroringKeyedCollection[
        KeyT,
        ResolvableKeyT,
        ValueT,
        MutableKeyedCollection[KeyT, ResolvableKeyT, ValueT, ResolvableValueT],
    ],
    MutableKeyedCollectionProxy[KeyT, ResolvableKeyT, ValueT, ResolvableValueT],
):
    """
    A mutable keyed collection proxy that raises custom errors.
    """

    @override
    def remove(self, *keys: KeyT | ResolvableKeyT) -> None:
        for key in keys:
            try:
                self._upstream.remove(key)
            except KeyError as error:
                raise self._key_error(error, key) from error

    @override
    def __delitem__(self, key: KeyT | ResolvableKeyT) -> None:
        try:
            del self._upstream[key]
        except KeyError as error:
            raise self._key_error(error, key) from error



"""
The Betty console.
"""

import argparse
import sys
from asyncio import CancelledError, run
from collections.abc import Iterable, Sequence
from enum import IntEnum
from typing import Any, cast, final, override

import rich  # noqa: F401
import rich_argparse

from betty.app import App
from betty.console.command import CommandDefinition, CommandFunction
from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.locale.localize import Localizer
from betty.user import Verbosity


@final
class SystemExitCode(IntEnum):
    """
    The exit codes used by the console.
    """

    OK = 0
    USER_QUIT = 1
    ERROR_CONSOLE_USAGE = 2
    ERROR_COMMAND_RUNTIME = 3
    ERROR_UNEXPECTED = 4


def _create_parser_class(*, localizer: Localizer) -> type[argparse.ArgumentParser]:
    class _ArgumentParser(argparse.ArgumentParser):
        def __init__(
            self,
            prog: str | None = None,
            usage: str | None = None,
            description: str | None = None,
            epilog: str | None = None,
            parents: Sequence[argparse.ArgumentParser] | None = None,
            formatter_class: type[
                argparse.HelpFormatter
            ] = rich_argparse.RawTextRichHelpFormatter,
            prefix_chars: str = "-",
            fromfile_prefix_chars: str | None = None,
            argument_default: Any = None,
            conflict_handler: str = "error",
            add_help: bool = True,
            allow_abbrev: bool = True,
            exit_on_error: bool = False,
            color: bool = False,
        ):
            super().__init__(
                prog,
                usage,
                description,
                epilog,
                parents or [],
                formatter_class,
                prefix_chars,
                fromfile_prefix_chars,
                argument_default,
                conflict_handler,
                False,
                allow_abbrev,
                exit_on_error,
            )
            self._positionals.title = localizer._("Positional arguments")
            self._optionals.title = localizer._("Options")
            self.add_argument(
                "--help",
                action="help",
                default=argparse.SUPPRESS,
                help=localizer._("Show this help message"),
            )

    return _ArgumentParser


def _create_formatter_class(*, localizer: Localizer) -> type[argparse.HelpFormatter]:
    class _HelpFormatter(rich_argparse.RawTextRichHelpFormatter):
        @override
        def _format_usage(
            self,
            usage: str | None,
            actions: Iterable[argparse.Action],
            groups: Iterable[argparse._MutuallyExclusiveGroup],
            prefix: str | None,
        ) -> str:
            if prefix is None:
                prefix = localizer._("Usage: ")
            return super()._format_usage(usage, actions, groups, prefix)

    return _HelpFormatter


async def _create_command_parser(
    app: App,
    subparsers: argparse._SubParsersAction,
    command_plugin: CommandDefinition,
    formatter_class: type[argparse.HelpFormatter],
) -> argparse.ArgumentParser:
    localizer = await app.localizer
    command = await app.factory.new(command_plugin.cls)
    command_parser: argparse.ArgumentParser = subparsers.add_parser(
        command.plugin().id,
        description=command.plugin().label.localize(localizer),
        exit_on_error=False,
        formatter_class=formatter_class,
    )
    command_parser.set_defaults(
        _command_func=await command.configure(command_parser),
        _verbosity=Verbosity.DEFAULT,
    )
    verbosity_group = command_parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        "-q",
        "--quiet",
        dest="_verbosity",
        action="store_const",
        const=Verbosity.QUIET,
        help=localizer._("Do not show any output, except error messages"),
    )
    verbosity_group.add_argument(
        "-v",
        "--verbose",
        dest="_verbosity",
        action="store_const",
        const=Verbosity.VERBOSE,
        help=localizer._("Also show detailed information messages"),
    )
    verbosity_group.add_argument(
        "-vv",
        "--more-verbose",
        dest="_verbosity",
        action="store_const",
        const=Verbosity.MORE_VERBOSE,
        help=localizer._("Also show debug messages"),
    )
    verbosity_group.add_argument(
        "-vvv",
        "--most-verbose",
        dest="_verbosity",
        action="store_const",
        const=Verbosity.MOST_VERBOSE,
        help=localizer._("Also show log messages"),
    )

    return command_parser


async def _create_list_commands_action_class(
    app: App, *, localizer: Localizer
) -> type[argparse.Action]:
    command_definitions = sorted(
        [x async for x in app.plugins[CommandDefinition]],
        key=lambda command_definition: command_definition.id,
    )

    class _ListCommandsAction(argparse.Action):
        def __init__(
            self,
            option_strings: Sequence[str],
            dest: str = argparse.SUPPRESS,
            default: Any = argparse.SUPPRESS,
            help: str | None = None,  # noqa: A002
            # Python 3.13 added the ``deprecated`` argument. For compatibility with all, allow it, but do not use it.
            deprecated: bool = False,
        ):
            super().__init__(
                option_strings=option_strings,
                dest=dest,
                default=default,
                nargs=0,
                help=help,
            )

        @override
        def __call__(
            self,
            parser: argparse.ArgumentParser,
            namespace: argparse.Namespace,
            values: str | Sequence[Any] | None,
            option_string: str | None = None,
        ):
            # Import rich locally because otherwise somehow Python would load betty.console.rich instead.
            import rich  # noqa: F811

            usage = localizer._("Usage: ")
            for (
                index,
                command_definition,  # noqa: F402
            ) in enumerate(command_definitions):
                if index != 0:
                    rich.print("")
                rich.print(
                    f"[dark_orange]{command_definition.label.localize(localizer)}:[/]"
                )
                rich.print(
                    f"  [cyan]{usage}[/][grey50]{parser.prog} {command_definition.id}[/]"
                )
                description = command_definition.description
                if description is not None:
                    rich.print(f"  {description.localize(localizer)}")
            raise SystemExit(SystemExitCode.OK)

    return _ListCommandsAction


async def _create_parser(app: App) -> argparse.ArgumentParser:
    localizer = await app.localizer
    argument_parser_class = _create_parser_class(localizer=localizer)
    formatter_class = _create_formatter_class(localizer=localizer)
    parser = argument_parser_class(
        exit_on_error=False, formatter_class=formatter_class, prog="betty"
    )
    parser.add_argument(
        "--commands",
        action=await _create_list_commands_action_class(app, localizer=localizer),
        default=argparse.SUPPRESS,
        help=localizer._("Show all available commands"),
    )
    subparsers = parser.add_subparsers(title=localizer._("Subcommands"))
    async for command_plugin in app.plugins[CommandDefinition]:
        await _create_command_parser(app, subparsers, command_plugin, formatter_class)
    return parser


async def main(app: App, args: Sequence[str]) -> None:
    """
    Launch Betty's console.

    :raises: SystemExit
    """
    parser = await _create_parser(app)
    try:
        namespace = parser.parse_args(args)
    except argparse.ArgumentError as error:
        await app.user.message_error(
            _("Invalid argument {argument}: {error}").format(
                argument=str(error.argument_name), error=error.message
            )
        )
        raise SystemExit(SystemExitCode.ERROR_CONSOLE_USAGE) from None
    try:
        command_func = cast(CommandFunction, namespace._command_func)
    except AttributeError:
        await app.user.message_error(
            _("It appears you called Betty without a command or arguments")
        )
        parser.print_help()
        raise SystemExit(SystemExitCode.ERROR_CONSOLE_USAGE) from None
    await app.user.set_verbosity(namespace._verbosity)
    try:
        await call_command_func(command_func, namespace)
        raise SystemExit(SystemExitCode.OK) from None
    except HumanFacingException as error:
        await app.user.message_error(error)
        raise SystemExit(SystemExitCode.ERROR_UNEXPECTED) from None
    except (CancelledError, KeyboardInterrupt):
        await app.user.message_information(_("Quitting..."))
        raise SystemExit(SystemExitCode.USER_QUIT) from None
    except Exception:
        await app.user.message_exception()
        raise SystemExit(SystemExitCode.ERROR_UNEXPECTED) from None


def main_standalone() -> None:
    """
    Launch Betty's console.

    This is a stand-alone entry point that will manage an event loop and Betty application.

    :raises: SystemExit
    """
    run(_main_standalone())


async def _main_standalone() -> None:
    async with await App.new_from_environment() as app:
        await main(app, sys.argv[1:])


async def call_command_func(
    command_func: CommandFunction, namespace: argparse.Namespace
) -> None:
    """
    Call a command function.
    """
    await command_func(
        **{
            name: value
            for name, value in vars(namespace).items()
            if not name.startswith("_")
        }
    )



"""
Provide demonstration site functionality.
"""

from __future__ import annotations

from asyncio import to_thread
from contextlib import suppress
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING, Self, final, override

from betty.copyright_notice.copyright_notices import Streetmix
from betty.extension import Extension, ExtensionDefinition
from betty.extension.demo.jobs import LoadAncestry
from betty.extension.deriver import Deriver
from betty.extension.http_api_doc import HttpApiDoc
from betty.extension.maps import Maps
from betty.extension.raspberry_mint import RaspberryMint
from betty.extension.spdx import Spdx
from betty.extension.trees import Trees
from betty.extension.wiki import Wiki
from betty.html import NavigationLink, NavigationLinkProvider
from betty.license import LicenseDefinition
from betty.license.licenses import spdx_license_id_to_license_id
from betty.locale.localizable.gettext import _
from betty.project import Project, generate
from betty.project.load import Loader, load
from betty.service.factory import Manufacturable
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from collections.abc import Sequence

    from betty.job import Context
    from betty.job.scheduler import Scheduler


async def generate_with_cleanup(
    project: Project, *, context: Context | None = None
) -> None:
    """
    Generate a demonstration site, and clean up the project directory on any errors.
    """
    if context:
        # Add a phantom value to the progress so it can never jump to 100% before we are entirely done here.
        await context.progress.add()

    if project.www_directory.exists():
        return
    await load(project, context=context)
    with suppress(FileNotFoundError):
        await to_thread(rmtree, project.directory)
    try:
        await generate.generate(project, context=context)
    except BaseException:
        with suppress(FileNotFoundError):
            await to_thread(rmtree, project.directory)
        raise

    if context:
        await context.progress.done()


@final
@ExtensionDefinition(
    "demo",
    label="Demo",
    depends_on={
        Deriver,
        HttpApiDoc,
        Maps,
        RaspberryMint,
        Spdx,
        Trees,
        Wiki,
    },
    assets_directory=Path(__file__).parent / "assets",
)
class Demo(NavigationLinkProvider, Loader, Manufacturable, Extension[Project]):
    """
    .. plugin:: extension:demo.
    """

    def __init__(self, *, project: Project):
        super().__init__()
        self._project = project

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(project=project)

    @override
    async def load(self, scheduler: Scheduler) -> None:
        licenses = self._project.plugins[LicenseDefinition]
        await scheduler.add(
            LoadAncestry(
                ancestry=self._project.ancestry,
                factory=self._project.factory,
                streetmix_copyright_notice=await self._project.factory.new(Streetmix),
                streetmix_license=await self._project.factory.new(
                    (
                        await licenses[
                            spdx_license_id_to_license_id("AGPL-3.0-or-later")
                        ]
                    ).cls
                ),
            )
        )

    @override
    def secondary_navigation_links(self) -> Sequence[NavigationLink]:
        return [
            NavigationLink(
                "https://github.com/bartfeenstra/betty", _("Find Betty on GitHub")
            ),
            NavigationLink(
                "https://betty.readthedocs.io/", _("Read the Betty documentation")
            ),
        ]



"""
Provide utilities for testing Betty or other software that depends on Betty.
"""

from typing import final


@final
class Counter:
    """
    An object that keeps track of how often it has been called.
    """

    def __init__(self):
        self._count = 0

    @property
    def count(self) -> int:
        """
        The call count.
        """
        return self._count

    def __call__(self) -> int:
        """
        Increment the counter.
        """
        self._count += 1
        return self._count



"""
Providing typing utilities.
"""

from __future__ import annotations

from typing import Any, final

from betty.docstring import append


def _should_mark(target: Any, key: str, /) -> bool:
    attr_name = f"_betty_typing_{key}"
    if hasattr(target, attr_name):
        return False
    setattr(target, attr_name, True)
    return True


def threadsafe[T](target: T, /) -> T:
    """
    Mark a target as thread-safe.
    """
    if _should_mark(target, "threadsafe"):
        target.__doc__ = append(
            target.__doc__ or "",
            "This is thread-safe, which means you can safely use this between different threads.",
        )
    return target


@final
class Void:
    """
    A sentinel that describes the absence of a value.

    Using this sentinel allows for actual values to be ``None``. Like ``None``,
    ``Void`` is only ever used through its type, and never instantiated.
    """

    def __new__():  # noqa: D102
        raise NotImplementedError


type VoidType = type[Void]



"""
Plugin factories.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, MutableSequence
from functools import cache
from json import dumps
from typing import TYPE_CHECKING, Any, Self, TypeVar, final, override

from betty.assertion import (
    AssertionChain,
    OptionalField,
    RequiredField,
    assert_mapping,
    assert_or,
    assert_record,
)
from betty.data import Data, DataDefinition
from betty.data.aggregate.record import PortableRecord, RecordDefinition
from betty.data.aggregate.record.object import AttrDefinition, ObjectDefinition
from betty.data.indicator.selector import Attr
from betty.locale.localizable.gettext import _
from betty.machine_name import MachineName
from betty.plugin import Plugin, PluginDefinition, ResolvableId, resolve_id
from betty.sample import Samplable, Sample, Samples, Size
from betty.typing import Void, VoidType

if TYPE_CHECKING:
    import builtins

    from betty.portable import PortableData
    from betty.service.level import ServiceLevel

_PluginManufacturerPluginT = TypeVar(
    "_PluginManufacturerPluginT", bound=Plugin, covariant=True
)
_PluginManufacturerPluginDefinitionT = TypeVar(
    "_PluginManufacturerPluginDefinitionT", bound=PluginDefinition
)


class PluginManufacturer[
    PluginManufacturerPluginDefinitionT: PluginDefinition,
    PluginManufacturerPluginT: Plugin,
](PortableRecord[Attr], Samplable, Data[RecordDefinition], ABC):
    """
    Configure a single plugin instance.
    """

    @final
    def __init__(
        self,
        plugin: ResolvableId[_PluginManufacturerPluginDefinitionT],
        data: Data | PortableData | VoidType = Void,
        /,
    ):
        super().__init__()
        self._plugin_id = resolve_id(plugin)
        self._plugin_data = data

    @final
    def __hash__(self):
        return hash(
            (
                self.type(),
                self.plugin_id,
                Void
                if self.plugin_data is Void
                else dumps(
                    self.plugin_data.data().porter.dump(self.plugin_data)
                    if isinstance(self.plugin_data, Data)
                    else self.plugin_data,
                ),
            )
        )

    @classmethod
    @abstractmethod
    def type(cls) -> builtins.type[_PluginManufacturerPluginDefinitionT]:
        """
        The type of plugin that can be manufactured.
        """

    @final
    @override
    @classmethod
    @cache
    def data(cls) -> ObjectDefinition[Self]:
        return ObjectDefinition(cls, label=cls.type().type().label)

    @final
    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.plugin_id, self.plugin_data) == (
            other.plugin_id,
            other.plugin_data,
        )

    @final
    @property
    @AttrDefinition(MachineName)
    def plugin_id(self) -> MachineName:
        """
        The plugin ID.
        """
        return self._plugin_id

    @final
    @property
    @AttrDefinition(DataDefinition(cls=object, label=_("Data")))
    def plugin_data(self) -> Data | PortableData | VoidType:
        """
        Get the plugin's own data.
        """
        return self._plugin_data

    @final
    @override
    @classmethod
    def load(cls, portable: PortableData, /) -> Self:
        record = assert_or(
            AssertionChain(MachineName.load)
            | (lambda plugin_id: {"plugin": plugin_id}),
            assert_record(
                RequiredField("plugin", MachineName.load),
                OptionalField("data"),
            ),
        )(portable)
        return cls(record["plugin"], record.get("data", Void))

    @final
    @override
    @classmethod
    def load_key(cls, portable: PortableData, key: Attr, portable_key: str, /) -> Self:
        return cls.load({**assert_mapping()(portable), "plugin": portable_key})

    @final
    def _dump_data(self, configuration: Data | PortableData) -> PortableData:
        if isinstance(configuration, Data):
            return configuration.data().porter.dump(configuration)
        return configuration

    @final
    @override
    def dump(self) -> PortableData:
        data = self.plugin_data
        if data is Void:
            return self._plugin_id
        return {
            "plugin": self._plugin_id,
            "data": self._dump_data(data),
        }

    @final
    @override
    def dump_key(self, key: Attr, /) -> tuple[str, PortableData]:
        return self.plugin_id, {} if self.plugin_data is Void else {
            "data": self._dump_data(self.plugin_data)
        }

    @final
    async def __call__(self, services: ServiceLevel, /) -> _PluginManufacturerPluginT:
        """
        Create a new instance of the configured plugin.
        """
        return await services.factory.new(
            (await services.plugins[self.type()][self.plugin_id]).cls,
            self.plugin_data,
        )

    @classmethod
    def resolve(
        cls,
        manufacturer: ResolvablePluginManufacturer[
            _PluginManufacturerPluginDefinitionT, _PluginManufacturerPluginT
        ],
    ) -> PluginManufacturer[
        _PluginManufacturerPluginDefinitionT, _PluginManufacturerPluginT
    ]:
        """
        Resolve a value to a plugin manufacturer.
        """
        if (
            isinstance(manufacturer, (PluginDefinition, str))
            or isinstance(manufacturer, type)
            and issubclass(manufacturer, Plugin)
        ):
            return cls(
                resolve_id(
                    manufacturer,  # ty:ignore[invalid-argument-type]
                )
            )
        return manufacturer

    @classmethod
    def resolve_sequence(
        cls,
        manufacturers: ResolvablePluginManufacturerSequence[
            _PluginManufacturerPluginDefinitionT, _PluginManufacturerPluginT
        ],
    ) -> MutableSequence[
        PluginManufacturer[
            _PluginManufacturerPluginDefinitionT, _PluginManufacturerPluginT
        ]
    ]:
        """
        Resolve a value to a sequence of plugin manufacturers.
        """
        if isinstance(manufacturers, PluginManufacturer):
            return [manufacturers]
        if (
            isinstance(manufacturers, (str, PluginDefinition))
            or isinstance(manufacturers, type)
            and issubclass(manufacturers, Plugin)
        ):
            return [cls.resolve(manufacturers)]
        return list(map(cls.resolve, manufacturers))  # ty:ignore[invalid-argument-type]

    @final
    @override
    @classmethod
    def samples(cls) -> Samples[Self]:
        return Samples(
            [
                lambda: Sample(
                    cls("my-first-plugin-id"),
                    label="Minimal",
                    size=Size.MINIMAL,
                ),
                lambda: Sample(
                    cls(
                        "my-first-plugin-id",
                        {
                            "configuration-key": "configuration-value",
                        },
                    ),
                    label="Full",
                    size=Size.FULL,
                ),
            ]
        )


type ResolvablePluginManufacturer[
    PluginDefinitionT: PluginDefinition,
    PluginT: Plugin,
] = ResolvableId[PluginDefinitionT] | PluginManufacturer[PluginDefinitionT, PluginT]


type ResolvablePluginManufacturerSequence[
    PluginDefinitionT: PluginDefinition,
    PluginT: Plugin,
] = (
    ResolvablePluginManufacturer[PluginDefinitionT, PluginT]
    | Iterable[ResolvablePluginManufacturer[PluginDefinitionT, PluginT]]
)



"""
The Betty Sphinx extension.
"""

from __future__ import annotations

from asyncio import run
from collections.abc import Callable, Iterable, Mapping, MutableSequence, Sequence
from functools import cmp_to_key
from textwrap import indent
from threading import Thread
from typing import TYPE_CHECKING, override

from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from sphinx.util.parsing import nested_parse_to_nodes

from betty.app import App
from betty.data import Data, OptionalDefinition
from betty.data.aggregate.record import RecordDefinition
from betty.definition.human_facing import HumanFacingDefinition
from betty.functools import Result
from betty.importlib import import_any
from betty.locale.localize import DEFAULT_LOCALIZER
from betty.plugin.dependent import DependentPluginDefinition
from betty.plugin.ordered import OrderedPluginDefinition
from betty.project import Project
from betty.serde import SerializerDefinition
from betty.service.factory import DataManufacturable
from betty.service.level import UNIVERSE

if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.util.typing import ExtensionMetadata

    from betty.machine_name import MachineName
    from betty.plugin import PluginDefinition
    from betty.serde import Serializer


type NodesLike = nodes.Node | Iterable[nodes.Node] | None


def _to_thread[T](target: Callable[[], T]) -> T:
    result = Result(target)
    thread = Thread(target=result)
    thread.start()
    thread.join()
    return result.result()


async def _get_plugins(plugin_type_id: str) -> Mapping[MachineName, PluginDefinition]:
    async with (
        App.new_isolated() as app,
        app,
        Project.new_isolated(app) as project,
        project,
    ):
        return {plugin.id: plugin async for plugin in project.plugins[plugin_type_id]}


def _cmp_formats(left: PluginDefinition, right: PluginDefinition) -> int:
    if left.id == "yaml":
        return -1
    if right.id == "yaml":
        return 1
    return -1 if left.id < right.id else 1


async def _get_serializers() -> Sequence[Serializer]:
    async with (
        App.new_isolated() as app,
        app,
        Project.new_isolated(app) as project,
        project,
    ):
        return [
            await project.factory.new(serializer.cls)
            for serializer in sorted(
                [x async for x in project.plugins[SerializerDefinition]],
                key=cmp_to_key(_cmp_formats),
            )
        ]


def _build_definition_list(
    definitions_nodes: Iterable[tuple[NodesLike, NodesLike]],
) -> nodes.Node:
    return nodes.definition_list(
        "",
        *[
            nodes.definition_list_item(
                "",
                nodes.term("", "", *_ensure_nodes(term_nodes)),
                nodes.definition("", *_ensure_nodes(definition_nodes)),
            )
            for term_nodes, definition_nodes in definitions_nodes
        ],
    )


def _ensure_nodes(nodes_like: NodesLike) -> Iterable[nodes.Node]:
    if nodes_like is None:
        return []
    if isinstance(nodes_like, nodes.Node):
        return [nodes_like]
    return nodes_like


class _PluginDirective(SphinxDirective):
    required_arguments = 1

    @override
    def run(self) -> list[nodes.Node]:
        # Right-strip periods to avoid D400 and D415 violations.
        argument = self.arguments[0].rstrip(".")
        try:
            plugin_type_id, plugin_id = argument.split(":")
        except ValueError:
            raise ValueError(
                f"The plugin directive requires a single argument that is a plugin type ID and a plugin ID, joined with a colon (:), but `{argument}` was given."
            ) from None
        plugins = _to_thread(lambda: run(_get_plugins(plugin_type_id)))
        plugin = plugins[plugin_id]
        return [
            self._build_summary(plugin),
            *self._build_metadata(plugin, plugins),
        ]

    def _build_summary(self, plugin: PluginDefinition) -> nodes.Node:
        summary_nodes, _ = self.parse_inline(
            f"The ``{plugin.id}`` :py:class:`{plugin.type().label.localize(DEFAULT_LOCALIZER).lower()} <{type(plugin).__module__}.{type(plugin).__qualname__}>` plugin."
        )
        if isinstance(plugin, HumanFacingDefinition):
            description = plugin.description
            if description:
                summary_nodes.append(nodes.Text(" "))
                summary_nodes.append(
                    nodes.Text(description.localize(DEFAULT_LOCALIZER))
                )
        return nodes.paragraph("", "", *summary_nodes)

    def _build_metadata(
        self, plugin: PluginDefinition, plugins: Mapping[MachineName, PluginDefinition]
    ) -> list[nodes.Node]:
        cls = plugin.cls
        if issubclass(cls, DataManufacturable):
            configuration_content = f":py:class:`{cls.new_data_cls().__name__} <{cls.new_data_cls().__module__}.{cls.new_data_cls().__qualname__}>`"
        else:
            configuration_content = "*not configurable*"
        content = f"""
.. list-table::
   :widths: 20 10
   :header-rows: 0

   * - Plugin ID
     - ``{plugin.id}``
   * - Class
     - :py:class:`{plugin.cls.__name__} <{plugin.cls.__module__}.{plugin.cls.__qualname__}>`
   * - Configuration
     - {configuration_content}
"""
        if isinstance(plugin, DependentPluginDefinition) and (
            depends_on_content := self._build_other_plugins_references(
                [plugins[plugin_id] for plugin_id in plugin.depends_on]
            )
        ):
            content += f"""
   * - Depends on
{depends_on_content}
"""
        if isinstance(plugin, OrderedPluginDefinition):
            if comes_before_content := self._build_other_plugins_references(
                [plugins[plugin_id] for plugin_id in plugin.comes_before]
            ):
                content += f"""
   * - Comes before
{comes_before_content}
"""
            if comes_after_content := self._build_other_plugins_references(
                [plugins[plugin_id] for plugin_id in plugin.comes_after]
            ):
                content += f"""
   * - Comes after
{comes_after_content}
"""
        return nested_parse_to_nodes(
            self.state,
            content,
            offset=self.content_offset,
        )

    def _build_other_plugins_references(
        self, plugins: Iterable[PluginDefinition]
    ) -> str:
        contents = [
            f":py:class:`{plugin.id} <{plugin.cls.__module__}.{plugin.cls.__qualname__}>`"
            for plugin in sorted(plugins, key=lambda plugin: plugin.id)
        ]
        if not contents:
            return ""
        if len(contents) == 1:
            return f"     - {contents[0]}"
        content = [f"     - * {contents[0]}"]
        for dependency_content in contents[1:]:
            content.append(f"       * {dependency_content}")
        return "\n".join(content)


class _PluginTypeDirective(SphinxDirective):
    required_arguments = 1

    @override
    def run(self) -> list[nodes.Node]:
        # Right-strip periods to avoid D400 and D415 violations.
        plugin_type_id = self.arguments[0].rstrip(".")
        plugin_type = UNIVERSE.plugins[plugin_type_id].type
        plugins = _to_thread(lambda: run(_get_plugins(plugin_type_id)))
        return [
            self._build_summary(plugin_type),
            *self._build_metadata(plugin_type),
            *self._build_builtin_plugins(plugin_type, plugins),
        ]

    def _build_summary(self, plugin_type: type[PluginDefinition]) -> nodes.Node:
        summary_node = nodes.paragraph(
            "",
            "",
            nodes.Text(
                f"The {plugin_type.type().label.localize(DEFAULT_LOCALIZER).lower()} plugin type."
            ),
        )
        description = plugin_type.type().description
        if description:
            summary_node.append(nodes.Text(" "))
            summary_node.append(nodes.Text(description.localize(DEFAULT_LOCALIZER)))
        return summary_node

    def _build_metadata(self, plugin_type: type[PluginDefinition]) -> list[nodes.Node]:
        return nested_parse_to_nodes(
            self.state,
            f"""
.. list-table::
   :widths: 20 10
   :header-rows: 0

   * - Plugin type ID
     - ``{plugin_type.type().id}``
   * - Definition
     - :py:class:`@{plugin_type.__name__}(...) <{plugin_type.__module__}.{plugin_type.__qualname__}>`
""",
            offset=self.content_offset,
        )

    def _build_builtin_plugins(
        self,
        plugin_type: type[PluginDefinition],
        plugins: Mapping[MachineName, PluginDefinition],
    ) -> list[nodes.Node]:
        return [
            nodes.paragraph(
                "",
                "",
                nodes.Text(
                    f"Built-in {plugin_type.type().label_plural.localize(DEFAULT_LOCALIZER).lower()}:"
                ),
            ),
            _build_definition_list(
                [
                    self._build_builtin_plugin_definition(plugin)
                    for plugin in sorted(plugins.values(), key=lambda plugin: plugin.id)
                ]
            ),
        ]

    def _build_builtin_plugin_definition(
        self, plugin: PluginDefinition
    ) -> tuple[NodesLike, NodesLike]:
        term_nodes, _ = self.parse_inline(
            f"{plugin.id} (:py:class:`{plugin.cls.__name__} <{plugin.cls.__module__}.{plugin.cls.__qualname__}>`)"
        )
        definition_nodes: MutableSequence[nodes.Node] | None = None
        if isinstance(plugin, HumanFacingDefinition):
            definition_nodes = [nodes.Text(plugin.label.localize(DEFAULT_LOCALIZER))]
            if plugin.description:
                definition_nodes.append(
                    nodes.Text(f": {plugin.description.localize(DEFAULT_LOCALIZER)}")
                )
        return term_nodes, definition_nodes


class _PluginTypesDirective(SphinxDirective):
    @override
    def run(self) -> list[nodes.Node]:
        return [
            _build_definition_list(
                [
                    self._build_builtin_plugin_type_definition(plugin_type.type)
                    for plugin_type in sorted(
                        UNIVERSE.plugins,
                        key=lambda plugin_type: plugin_type.type.type().label.localize(
                            DEFAULT_LOCALIZER
                        ),
                    )
                ]
            ),
        ]

    def _build_builtin_plugin_type_definition(
        self, plugin_type: type[PluginDefinition]
    ) -> tuple[NodesLike, NodesLike]:
        term_nodes, _ = self.parse_inline(
            f":py:class:`{plugin_type.type().label.localize(DEFAULT_LOCALIZER)} <{plugin_type.__module__}.{plugin_type.__qualname__}>` (``{plugin_type.type().id}``)"
        )
        description = plugin_type.type().description
        if description:
            return term_nodes, nodes.Text(description.localize(DEFAULT_LOCALIZER))
        return term_nodes, None


class _DataDirective(SphinxDirective):
    required_arguments = 1

    @override
    def run(self) -> list[nodes.Node]:
        # Right-strip periods to avoid D400 and D415 violations.
        cls_name = self.arguments[0].rstrip(".")
        cls = import_any(cls_name)
        assert issubclass(cls, Data)
        data = cls.data()
        content = ""

        if isinstance(data, RecordDefinition) and data.fields:
            content += """
Data
====
"""
            for field in sorted(
                data.fields,
                key=lambda field: (
                    isinstance(field.subject, OptionalDefinition),
                    field.selector.element,
                ),
            ):
                primary_label = field.data.label if field.label is None else field.label
                content += f"""
                
``{field.selector.element}`` :sup:`{"optional" if isinstance(field.data, OptionalDefinition) else "required"}`

    **{primary_label.localize(DEFAULT_LOCALIZER)}**
"""
                primary_description = (
                    field.data.description
                    if field.description is None
                    else field.description
                )
                if primary_description is not None:
                    content += f"""
    {primary_description.localize(DEFAULT_LOCALIZER)}
"""
                content += f"""

    Value: :py:class:`{field.data.cls.__name__} <{field.data.cls.__module__}.{field.data.cls.__qualname__}>`
"""
                if field.label is not None:
                    content += f"""
    *{field.data.label.localize(DEFAULT_LOCALIZER)}*
"""
                if field.description is not None and field.data.description is not None:
                    content += f"""
    *{field.data.description.localize(DEFAULT_LOCALIZER)}*
"""

        samples = list(data.samples)
        if samples:
            examples_label = "Examples" if len(samples) > 1 else "Example"
            content += f"""
{examples_label}
{"".join(["=" * len(examples_label)])}

"""
            serializers = _to_thread(lambda: run(_get_serializers()))
            for sample in samples:
                example_content = ""
                if len(samples) > 1:
                    example_label = sample.label.localize(DEFAULT_LOCALIZER)
                    example_content += f"""
{example_label}
{"".join(["-" * len(example_label)])}
"""
                example_content += """
.. tab-set::

"""
                portable = data.dump(sample.subject)
                for serializer in serializers:
                    serialized = serializer.dump(portable)
                    example_content += f"""
   .. tab-item:: {serializer.plugin().label.localize(DEFAULT_LOCALIZER)}

      .. code-block:: {serializer.plugin().id}

{indent(serialized, " " * 10)}
"""
                content += example_content

        return nested_parse_to_nodes(
            self.state,
            content,
            offset=self.content_offset,
        )


def setup(app: Sphinx) -> ExtensionMetadata:
    """
    Implement Sphinx's extension setup.
    """
    app.add_directive("data", _DataDirective)
    app.add_directive("plugin", _PluginDirective)
    app.add_directive("plugin_type", _PluginTypeDirective)
    app.add_directive("plugin_types", _PluginTypesDirective)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
Provide the Raspberry Mint theme.
"""

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Self, final, override

from betty.extension import ExtensionDefinition
from betty.extension._theme import jinja_filters
from betty.extension.maps import Maps
from betty.extension.raspberry_mint.data import RaspberryMintConfiguration
from betty.extension.trees import Trees
from betty.extension.webpack import Webpack
from betty.extension.webpack.build import EntryPointProvider
from betty.jinja import Filters, JinjaProvider
from betty.model import EntityDefinition
from betty.project import Project
from betty.project.generate import Generator
from betty.service.factory import DataManufacturable, Manufacturable
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from collections.abc import Sequence

    from betty.job.scheduler import Scheduler


@final
@ExtensionDefinition(
    "raspberry-mint",
    label="Raspberry Mint",
    depends_on={Webpack},
    comes_before={
        Maps,
        Trees,
    },
    theme=True,
    assets_directory=Path(__file__).parent / "assets",
)
class RaspberryMint(
    DataManufacturable[RaspberryMintConfiguration],
    Manufacturable,
    JinjaProvider,
    Generator,
    EntryPointProvider[Project],
):
    """
    .. plugin:: extension:raspberry-mint.

    .. important::
        This extension requires :ref:`Node.js <installation-requirements-nodejs>`.

    Regions
    -------

    Raspberry Mint provides the following regions contents may be configured for:

    - ``front-page-content``
      The main content for the front page.
    - ``front-page-summary``
      The page summary for the front page.
    - ``entity-page-content``
      The page content region for entity pages.
    - ``entity-page-content--{entity_type_id}``
      The page content region for entity pages of a specific public-facing entity type, where ``{entity_type_id}`` is the
      entity type ID. If no content is assigned to this region for an entity type, ``entity-page-content`` is used instead.

    Templating
    ----------

    Filters
    ^^^^^^^

    - :py:func:`associated_file_references <betty.extension._theme.associated_file_references>`
    - :py:func:`person_descendant_families <betty.extension._theme.person_descendant_families>`
    - :py:func:`person_timeline_events <betty.extension._theme.person_timeline_events>`

    """

    def __init__(
        self,
        *,
        project: Project,
        configuration: RaspberryMintConfiguration | None = None,
    ):
        super().__init__()
        self._project = project
        self._configuration = (
            RaspberryMintConfiguration() if configuration is None else configuration
        )

    @override
    @classmethod
    def new_data_cls(cls) -> type[RaspberryMintConfiguration]:
        return RaspberryMintConfiguration

    @property
    def configuration(self) -> RaspberryMintConfiguration:
        """
        The configuration.
        """
        return self._configuration

    @override
    @classmethod
    @require_project
    async def new(
        cls,
        project: Project,
        data: RaspberryMintConfiguration | None = None,
        /,
    ) -> Self:
        return cls(configuration=data, project=project)

    @override
    async def generate(self, scheduler: Scheduler) -> None:
        from betty.extension.raspberry_mint.jobs import (
            _GenerateLogo,
            _GenerateSearchIndex,
            _GenerateWebmanifest,
        )

        await scheduler.add(
            _GenerateLogo(project=self._project),
            _GenerateSearchIndex(project=self._project),
            _GenerateWebmanifest(project=self._project),
        )

    @override
    @classmethod
    def webpack_entry_point_directory_path(cls) -> Path:
        return Path(__file__).parent / "webpack"

    @override
    def webpack_entry_point_cache_keys(self) -> Sequence[str]:
        return (
            self._project.configuration.root_path,
            self._configuration.primary_color,
            self._configuration.secondary_color,
            self._configuration.tertiary_color,
        )

    @override
    @property
    def filters(self) -> Filters:
        return jinja_filters(self._project)

    @property
    async def regions(self) -> set[str]:
        """
        The available regions.
        """
        return {
            "front-page-content",
            "front-page-summary",
            "entity-page-content",
            *{
                f"entity-page-content--{entity_type.id}"
                async for entity_type in self._project.plugins[EntityDefinition]
                if entity_type.public_facing
            },
        }


@final
class ColorStyle(Enum):
    """
    The available color styles.
    """

    LIGHT = "light"
    """
    A light style with a white background.
    """

    LIGHT_SECONDARY = "light-secondary"
    """
    A light style with a light shade of the secondary color for the background.
    """

    LIGHT_CONTRAST = "light-contrast"
    """
    A light style with a light shade of gray for the background.
    """

    DARK = "dark"
    """
    A dark style with a black background.
    """

    DARK_SECONDARY = "dark-secondary"
    """
    A dark style with a dark shade of the secondary color for the background.
    """


@final
class Breakpoint(Enum):
    """
    The theme's breakpoints.
    """

    XS = "xs"
    SM = "sm"
    MD = "md"
    LG = "lg"
    XL = "xl"
    XXL = "xxl"


@final
class JustifyContent(Enum):
    """
    How to justify content.
    """

    START = "start"
    END = "end"
    CENTER = "center"
    BETWEEN = "between"
    AROUND = "around"
    EVENLY = "evenly"


SINGLE_COLUMN_TEXT_WIDTH = {
    Breakpoint.XS: 12,
    Breakpoint.LG: 11,
    Breakpoint.XL: 10,
    Breakpoint.XXL: 9,
}



"""
File path tools.
"""

from pathlib import Path
from typing import final

from betty.assertion import assert_path
from betty.data import DataDefinition
from betty.locale.localizable.gettext import _
from betty.portable import CallbackPorter


@final
class FilePathDefinition(DataDefinition[Path]):
    """
    A file path definition.
    """

    def __init__(self):
        super().__init__(
            cls=Path,
            label=_("File path"),
            porter=CallbackPorter[Path](assert_path(), str),
        )



"""
Provide the Command Line Interface.
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, final

from betty.definition.human_facing import HumanFacingDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition

if TYPE_CHECKING:
    import argparse

type CommandFunction = Callable[..., Awaitable[None]]


class Command(Plugin["CommandDefinition"]):
    """
    A console command plugin.
    """

    @abstractmethod
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        """
        Configure the command.

        :return: The command function, which is an async callable that returns ``None`` and takes all parser arguments
            as keyword arguments.
        """


@final
@PluginTypeDefinition(
    "command",
    label=_("Command"),
    label_plural=_("Commands"),
    label_countable=ngettext("{count} command", "{count} commands"),
)
class CommandDefinition(HumanFacingDefinition, PluginDefinition[Command]):
    """
    .. plugin_type:: command.
    """



"""
Provide the Documentation API.
"""

import multiprocessing
from asyncio import to_thread
from contextlib import AsyncExitStack
from pathlib import Path
from shutil import copytree
from typing import final, override

from aiofiles.os import makedirs
from sphinx.application import Sphinx
from sphinx.ext.autodoc import MethodDocumenter

from betty import serve
from betty.dirs import ROOT_DIRECTORY_PATH
from betty.exception import HumanFacingException
from betty.serve import NoPublicUrlBecauseServerNotStartedError, Server
from betty.user import User, Verbosity


async def _ensure_www_directory(
    output_directory_path: Path, cache_directory_path: Path, *, user: User
) -> None:
    if not output_directory_path.exists():
        await _build(output_directory_path, cache_directory_path, user=user)


async def _build(
    output_directory_path: Path, cache_directory_path: Path, *, user: User
) -> None:
    await makedirs(output_directory_path, exist_ok=True)
    # sphinx-apidoc must output to the documentation directory, but because we do not want
    # to 'pollute' that with generated files that must not be committed, do our work in a
    # dedicated cache directory.
    source_directory_path = cache_directory_path / "source"
    await to_thread(
        copytree, ROOT_DIRECTORY_PATH / "documentation", source_directory_path
    )
    sphinx_app = Sphinx(
        buildername="dirhtml",
        confdir=str(source_directory_path),
        doctreedir=str(cache_directory_path / ".doctrees"),
        outdir=str(output_directory_path),
        parallel=multiprocessing.cpu_count(),
        srcdir=str(source_directory_path),
        verbosity=9 if user.verbosity is Verbosity.MOST_VERBOSE else 0,
        warningiserror=True,
    )
    # Work around a bug in Sphinx where MethodDocumenter.can_document_member would erroneously consider our descriptors
    # as methods resulting in errors being raised because said descriptors are not callable and do not have a signature.
    original_can_document_member = MethodDocumenter.can_document_member
    MethodDocumenter.can_document_member = (  # ty:ignore[invalid-assignment]
        lambda member, membername, isattr, parent: (
            original_can_document_member(member, membername, isattr, parent)
            and callable(member)
        )
    )
    try:
        sphinx_app.build()
    finally:
        MethodDocumenter.can_document_member = original_can_document_member  # ty:ignore[invalid-assignment]
    if sphinx_app.statuscode != 0:
        raise HumanFacingException("Sphinx failed.")


@final
class DocumentationServer(Server):
    """
    Serve the documentation site.
    """

    def __init__(self, cache_directory_path: Path, *, user: User):
        super().__init__(user=user)
        self._cache_directory_path = cache_directory_path
        self._server: Server | None = None
        self._exit_stack = AsyncExitStack()

    @override
    @property
    def public_url(self) -> str:
        if self._server is not None:
            return self._server.public_url
        raise NoPublicUrlBecauseServerNotStartedError()

    @override
    async def start(self) -> None:
        www_directory_path = self._cache_directory_path / "www"
        await _ensure_www_directory(
            www_directory_path, self._cache_directory_path / "cache", user=self._user
        )
        self._server = serve.BuiltinServer(www_directory_path, user=self._user)
        await self._exit_stack.enter_async_context(self._server)

    @override
    async def stop(self) -> None:
        await self._exit_stack.aclose()



"""
Provide the OpenAPI specification.
"""

from betty import about
from betty.locale.localize import DEFAULT_LOCALIZER
from betty.model import EntityDefinition
from betty.portable import PortableMapping
from betty.project import Project
from betty.project.schema import ProjectSchema
from betty.string import kebab_case_to_lower_camel_case


class Specification:
    """
    Build OpenAPI specifications.
    """

    def __init__(self, project: Project):
        self._project = project

    async def build(self) -> PortableMapping:
        """
        Build the OpenAPI specification.
        """
        url_generator = await self._project.url_generator
        specification_paths: PortableMapping[PortableMapping] = {}
        specification: PortableMapping = {
            "openapi": "3.1.0",
            "servers": [
                {
                    "url": url_generator.generate("betty-static:///", absolute=True),
                }
            ],
            "info": {
                "title": "Betty",
                "version": about.VERSION_LABEL,
            },
            "paths": specification_paths,
            "components": {
                "responses": {
                    "401": {
                        "description": "Unauthorized",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": await ProjectSchema.def_url(
                                        self._project, "errorResponse"
                                    ),
                                },
                            },
                        },
                    },
                    "403": {
                        "description": "Forbidden",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": await ProjectSchema.def_url(
                                        self._project, "errorResponse"
                                    ),
                                },
                            },
                        },
                    },
                    "404": {
                        "description": "Not found",
                        "content": {
                            "application/json": {
                                "schema": {
                                    "$ref": await ProjectSchema.def_url(
                                        self._project, "errorResponse"
                                    ),
                                },
                            },
                        },
                    },
                },
                "parameters": {
                    "id": {
                        "name": "id",
                        "in": "path",
                        "required": True,
                        "description": "The ID for the resource to retrieve.",
                        "schema": {
                            "type": "string",
                        },
                    },
                },
                "schemas": {
                    "betty": {
                        "$ref": await ProjectSchema.url(self._project),
                    },
                },
            },
        }

        # Add entity operations.
        async for entity_type in self._project.plugins[EntityDefinition]:
            if not entity_type.public_facing:
                continue
            await entity_type.cls.linked_data_schema(self._project)
            if self._project.configuration.clean_urls:
                collection_path = f"/{entity_type.id}/"
                single_path = f"/{entity_type.id}/{{id}}/"
            else:
                collection_path = f"/{entity_type.id}/index.json"
                single_path = f"/{entity_type.id}/{{id}}/index.json"
            entity_type_label = entity_type.label.localize(DEFAULT_LOCALIZER)
            specification_paths[collection_path] = {
                "get": {
                    "summary": f"Retrieve the collection of {entity_type_label} entities.",
                    "responses": {
                        "200": {
                            "description": f"The collection of {entity_type_label} entities.",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "$ref": await ProjectSchema.def_url(
                                            self._project,
                                            f"{kebab_case_to_lower_camel_case(entity_type.id)}EntityCollectionResponse",
                                        ),
                                    },
                                },
                            },
                        },
                    },
                    "tags": [entity_type_label],
                },
            }
            specification_paths[single_path] = {
                "get": {
                    "summary": f"Retrieve a single {entity_type_label} entity.",
                    "responses": {
                        "200": {
                            "description": f"The {entity_type_label} entity.",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "$ref": await ProjectSchema.def_url(
                                            self._project,
                                            f"{kebab_case_to_lower_camel_case(entity_type.id)}Entity",
                                        ),
                                    },
                                },
                            },
                        },
                    },
                    "tags": [entity_type_label],
                },
            }

        # Add default behavior to all requests.
        for path_specification in specification_paths.values():
            path_specification["get"]["responses"].update(
                {
                    "401": {
                        "$ref": "#/components/responses/401",
                    },
                    "403": {
                        "$ref": "#/components/responses/403",
                    },
                    "404": {
                        "$ref": "#/components/responses/404",
                    },
                }
            )

        return specification



"""
Provide the Locale API.
"""

from __future__ import annotations

from collections.abc import Sequence
from contextlib import suppress
from typing import TYPE_CHECKING, Any, override

from babel import Locale
from babel.core import UnknownLocaleError

import betty.dirs

if TYPE_CHECKING:
    from collections.abc import Sequence


_LOCALE_DIRECTORY_PATH = betty.dirs.ASSETS_DIRECTORY_PATH / "locale"


DEFAULT_LOCALE = Locale("en", "US")
"""
Betty's default locale (US English).
"""

DEFAULT_LOCALE_TAG = "en-US"
"""
The `IETF BCP 47 <https://tools.ietf.org/html/bcp47>`_ language tag for Betty's default locale (US English).
"""


type ResolvableLocale = Locale | str
"""
A locale or a locale identifier.
"""


def resolve_locale(locale: ResolvableLocale, /) -> Locale:
    """
    Ensure that the given value is a locale.

    :raises betty.locale.InvalidLocale: Raised if the given identifier is not a valid locale.
    :raises betty.locale.LocaleNotFoundError: Raised if the given locale cannot be found.
    """
    if isinstance(locale, Locale):
        return locale
    return from_language_tag(locale)


def to_language_tag(locale: Locale | None, /) -> str:
    """
    Formats a locale as an `IETF BCP 47 <https://tools.ietf.org/html/bcp47>`_ language tag.
    """
    if locale is None:
        return "und"
    return "-".join(
        part
        for part in [
            locale.language,
            locale.script,
            locale.territory,
        ]
        if part
    )


def from_language_tag(locale: str, /) -> Locale:
    """
    Get a locale from its `IETF BCP 47 <https://tools.ietf.org/html/bcp47>`_ language tag.

    :raises betty.locale.InvalidLocale: Raised if the given identifier is not a valid locale.
    :raises betty.locale.LocaleNotFoundError: Raised if the given locale cannot be found.
    """
    try:
        return Locale.parse(locale, sep="-")
    except ValueError:
        from betty.locale.error import InvalidLocale

        raise InvalidLocale(locale) from None
    except UnknownLocaleError:
        from betty.locale.error import UnknownLocale

        raise UnknownLocale(locale) from None


def negotiate_locale(
    preferred_locales: Locale | Sequence[Locale],
    available_locales: Sequence[Locale],
    /,
) -> Locale | None:
    """
    Negotiate the preferred locale from a sequence.
    """
    if isinstance(preferred_locales, Locale):
        preferred_locales = [preferred_locales]
    preferred_locale_babel_identifiers = list(map(str, preferred_locales))  # ty:ignore[invalid-argument-type]
    available_locale_babel_identifiers = list(map(str, available_locales))
    negotiated_locale = Locale.negotiate(
        preferred_locale_babel_identifiers, available_locale_babel_identifiers
    )
    if negotiated_locale is not None:
        return negotiated_locale
    return Locale.negotiate(
        [
            (
                babel_identifier.split("_")[0]
                if "_" in babel_identifier
                else babel_identifier
            )
            for babel_identifier in preferred_locale_babel_identifiers
        ],
        available_locale_babel_identifiers,
    )


def plural_tags(locale: Locale) -> Sequence[str]:
    """
    Get a locale's plural tags.
    """
    tags = list(locale.plural_form.tags)
    if "other" not in tags:
        tags.append("other")
    return tags


class HasLocale:
    """
    A resource that has a locale, e.g. contains information in a specific locale.
    """

    def __init__(
        self, *args: Any, locale: ResolvableLocale | None = None, **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        self._locale = None if locale is None else resolve_locale(locale)

    @property
    def locale(self) -> Locale | None:
        """
        The locale the data in this instance is in.
        """
        return self._locale


def negotiate_has_locales(
    preferred_locales: Locale | Sequence[Locale],
    has_locales: Sequence[HasLocale],
) -> HasLocale | None:
    """
    Negotiate the preferred value from a sequence.
    """
    negotiated_locale = negotiate_locale(
        preferred_locales,
        [
            has_locale.locale
            for has_locale in has_locales
            if has_locale.locale is not None
        ],
    )
    if negotiated_locale is not None:
        for has_locale in has_locales:
            if has_locale.locale == negotiated_locale:
                return has_locale
    for has_locale in has_locales:
        if has_locale.locale is None:
            return has_locale
    with suppress(IndexError):
        return has_locales[0]
    return None


class HasLocaleStr(HasLocale, str):
    """
    A string that has a locale.
    """

    __slots__ = "_locale"

    @override
    def __new__(cls, string: str, *, locale: Locale | None = None):
        new = super().__new__(cls, string)
        new._locale = locale
        return new

    def __init__(self, string: str, *, locale: Locale | None = None):
        pass



"""
Errors for the localizable API.
"""

from __future__ import annotations

from typing import final

from betty.locale.error import LocaleError


@final
class MissingPluralPlaceholder(LocaleError):
    """
    Raised when a plural translation is missing a placeholder.
    """


@final
class MissingPluralTag(LocaleError):
    """
    Raised when a countable localizable is missing a plural tag.
    """


@final
class InvalidPluralTag(LocaleError):
    """
    Raised when a countable localizable defines an invalid plural tag.
    """



"""
Provide concrete gender implementations.
"""

from typing import final

from betty.classtools import Singleton
from betty.gender import Gender, GenderDefinition
from betty.locale.localizable.gettext import _, ngettext


@final
@GenderDefinition(
    "man",
    label=_("Man"),
    label_plural=_("Men"),
    label_countable=ngettext("{count} man", "{count} men"),
)
class Man(Gender):
    """
    .. plugin:: gender:man.
    """


@final
@GenderDefinition(
    "non-binary",
    label=_("Non-binary person"),
    label_plural=_("Non-binary people"),
    label_countable=ngettext("{count} non-binary person", "{count} non-binary people"),
)
class NonBinary(Gender):
    """
    .. plugin:: gender:non-binary.
    """


@final
@GenderDefinition(
    "unknown",
    label=_("Person of unknown gender"),
    label_plural=_("People of unknown gender"),
    label_countable=ngettext(
        "{count} person of unknown gender", "{count} people of unknown gender"
    ),
)
class Unknown(Gender, Singleton):
    """
    .. plugin:: gender:unknown.
    """


@final
@GenderDefinition(
    "woman",
    label=_("Woman"),
    label_plural=_("Women"),
    label_countable=ngettext("{count} woman", "{count} women"),
)
class Woman(Gender):
    """
    .. plugin:: gender:woman.
    """



"""
Plugin configuration properties.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from betty.collection.keyed.adapter import MutableKeyedCollectionAdapter
from betty.collection.sequence import (
    MutableResolvedSequence,
)
from betty.data.aggregate.collection.keyed import KeyedCollectionDefinition
from betty.data.indicator.selector import Attr
from betty.plugin import Plugin, PluginDefinition
from betty.plugin.data import PluginManufacturerSequenceDefinition
from betty.plugin.factory import (
    PluginManufacturer,
    ResolvablePluginManufacturer,
    ResolvablePluginManufacturerSequence,
)
from betty.property.collection.keyed import KeyedCollectionProperty
from betty.property.collection.sequence import SequenceProperty

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable
    from betty.plugin.data import PluginDefinitionConfiguration


@final
class PluginManufacturerSequenceProperty[
    PluginDefinitionT: PluginDefinition,
    PluginT: Plugin,
](
    SequenceProperty[
        MutableResolvedSequence[
            PluginManufacturer[PluginDefinitionT, PluginT],
            ResolvablePluginManufacturer[PluginDefinitionT, PluginT],
        ],
        ResolvablePluginManufacturerSequence[PluginDefinitionT, PluginT],
    ]
):
    """
    A property containing a sequence of :py:class:`betty.plugin.factory.PluginManufacturer`.
    """

    def __init__(
        self,
        manufacturer: type[PluginManufacturer[PluginDefinitionT, PluginT]],
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            PluginManufacturerSequenceDefinition(manufacturer),
            label=label,
            description=description,
            resolver=manufacturer.resolve_sequence,
        )


@final
class PluginDefinitionConfigurationsProperty[PluginDefinitionT: PluginDefinition](
    KeyedCollectionProperty
):
    """
    A property containing a :py:class:`betty.collection.keyed.KeyedCollection` of :py:class:`betty.plugin.data.PluginDefinitionConfiguration`.
    """

    def __init__(
        self,
        plugin_type: type[PluginDefinitionT],
        item: type[PluginDefinitionConfiguration[PluginDefinitionT]],
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            KeyedCollectionDefinition(
                value=item,
                label=plugin_type.type().label_plural,
                key=Attr("id"),
                factory=lambda: MutableKeyedCollectionAdapter(key=lambda item: item.id),
            ),
            label=label,
            description=description,
            omit_load=True,
            omit_dump=lambda data: not len(data),
        )



"""Integrate Betty with `Swagger UI <https://swagger.io/tools/swagger-ui>`_."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, final, override

from betty.extension import ExtensionDefinition
from betty.extension.webpack import Webpack
from betty.extension.webpack.build import EntryPointProvider
from betty.html import NavigationLink, NavigationLinkProvider
from betty.locale.localizable.gettext import _

if TYPE_CHECKING:
    from collections.abc import Sequence


@final
@ExtensionDefinition(
    "http-api-doc",
    label="HTTP API Documentation",
    description=_(
        "Display the HTTP API documentation in a user-friendly way using Swagger UI."
    ),
    depends_on={Webpack},
    assets_directory=Path(__file__).parent / "assets",
)
class HttpApiDoc(EntryPointProvider, NavigationLinkProvider):
    """
    .. plugin:: extension:http-api-doc.
    """

    # Provide an initializer without arguments so the factory can call it.
    def __init__(self):
        super().__init__()

    @override
    @classmethod
    def webpack_entry_point_directory_path(cls) -> Path:
        return Path(__file__).parent / "webpack"

    @override
    def webpack_entry_point_cache_keys(self) -> Sequence[str]:
        return ()

    @override
    def secondary_navigation_links(self) -> Sequence[NavigationLink]:
        return [
            NavigationLink("betty-static:///api/index.html", _("API documentation")),
        ]



"""
Data for the Wikipedia extension.
"""

from typing import final

from betty.data import Data, Sample
from betty.data.aggregate.record.object import ObjectDefinition
from betty.data.bool import BoolDefinition
from betty.locale.localizable.gettext import _
from betty.property import Optional, Property
from betty.sample import Size


@final
@ObjectDefinition(
    label=_("Wiki extension configuration"),
    samples=[
        lambda: Sample(WikiConfiguration(), label="Minimal", size=Size.MINIMAL),
        lambda: Sample(
            WikiConfiguration(populate_images=False), label="Full", size=Size.FULL
        ),
    ],
)
class WikiConfiguration(Data):
    """
    Configuration for the :py:class:`betty.extension.wiki.Wiki` extension.

    .. data:: betty.extension.wiki.data:WikiConfiguration
    """

    populate_images = Optional(
        Property(
            BoolDefinition(
                label=_("Populate images"),
                description=_(
                    "Whether to download additional images found through Wikipedia links in the ancestry"
                ),
            ),
            omit_load=True,
            omit_dump=lambda data: data is True,
        )
    )
    """
    Whether to populate entities with Wikimedia images after loading ancestries.
    """

    def __init__(self, *, populate_images: bool = True):
        self.populate_images = populate_images



"""
Job execution using async/await.
"""

from __future__ import annotations

from asyncio import CancelledError, Task, as_completed, get_running_loop
from contextlib import suppress
from typing import TYPE_CHECKING, final, override

from betty.job.executor import Executor
from betty.job.scheduler import Cancelled, Completed

if TYPE_CHECKING:
    from betty.job.scheduler import Scheduler


@final
class AsyncExecutor(Executor):
    """
    A job executor using async/await.
    """

    def __init__(self, scheduler: Scheduler, *, concurrency: int = 1):
        assert concurrency > 0
        self._scheduler = scheduler
        self._concurrency = concurrency
        self._working = False
        self._tasks: set[Task[None]] = set()

    @override
    async def start(self) -> None:
        if self._working:
            return
        self._working = True
        event_loop = get_running_loop()
        for _ in range(self._concurrency):
            task = event_loop.create_task(self._run_job())
            self._tasks.add(task)

    async def _run_job(self) -> None:
        with suppress(Completed, CancelledError):
            try:
                while self._working:
                    batch = await self._scheduler.get()
                    await batch()
            except Cancelled:
                await self.cancel()

    @override
    async def cancel(self) -> None:
        self._working = False
        for task in self._tasks:
            task.cancel()
        for task in as_completed(self._tasks):
            with suppress(CancelledError):
                await task

    @override
    async def complete(self) -> None:
        if not self._working:
            return
        for task in as_completed(self._tasks):
            with suppress(CancelledError):
                await task
        self._working = False



"""
Test utilities for :py:mod:`betty.json.linked_data`.
"""

from __future__ import annotations

from collections.abc import Awaitable, Callable, MutableMapping, MutableSequence
from typing import TYPE_CHECKING

from betty.app import App
from betty.json.schema import Schema
from betty.portable import PortableData
from betty.project import Project

if TYPE_CHECKING:
    from betty.json.linked_data import (
        LinkedDataDumpableWithSchema,
        LinkedDataDumper,
    )


async def assert_dumps_linked_data[PortableDataT: PortableData](
    sut: LinkedDataDumpableWithSchema[Schema, PortableDataT],
) -> PortableDataT:
    """
    Dump an object's linked data and assert it is valid.
    """
    return await assert_linked_data_dump(sut.linked_data_schema, sut.dump_linked_data)


async def assert_dumps_linked_data_for[PortableDataT: PortableData, T](
    sut: LinkedDataDumper[T, Schema, PortableDataT], target: T
) -> PortableDataT:
    """
    Dump an object's linked data and assert it is valid.
    """

    async def _dump(project: Project) -> PortableDataT:
        return await sut.dump_linked_data_for(project, target)

    return await assert_linked_data_dump(sut.linked_data_schema_for, _dump)


async def assert_linked_data_dump[PortableDataT: PortableData](
    schema: Callable[[Project], Awaitable[Schema]] | Schema,
    portable: Callable[[Project], Awaitable[PortableDataT]] | PortableDataT,
) -> PortableDataT:
    """
    Assert that dumped linked data is valid against a schema.
    """
    async with (
        App.new_isolated() as app,
        app,
        Project.new_isolated(app) as project,
        project,
    ):
        actual = await portable(project) if callable(portable) else portable  # ty:ignore[call-top-callable]

        # Validate the raw dump.
        sut_schema = schema if isinstance(schema, Schema) else await schema(project)
        sut_schema.validate(actual)

        # Normalize the dump after validation (so we are assured it is absolutely valid),
        # but before returning, so calling code can use simpler comparisons.
        return _normalize(actual)


def _normalize[PortableDataT: PortableData](portable: PortableDataT) -> PortableDataT:
    if isinstance(portable, MutableMapping):
        return {
            key: _normalize(value)
            for key, value in portable.items()
            if not key.startswith("$")
        }  # ty:ignore[invalid-return-type]
    if isinstance(portable, MutableSequence) and not isinstance(portable, str):
        return list(
            map(
                _normalize,
                portable,  # ty:ignore[invalid-argument-type]
            )
        )  # ty:ignore[invalid-return-type]
    return portable



"""
Data types to reference files on disk.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from betty.ancestry.file import File
from betty.locale.localizable.gettext import _, ngettext
from betty.model import Entity, EntityDefinition
from betty.model.association import BidirectionalToOne, ToOneAssociate

if TYPE_CHECKING:
    from betty.ancestry.has_file_references import HasFileReferences
    from betty.image import FocusArea


@final
@EntityDefinition(
    "file-reference",
    label=_("File reference"),
    label_plural=_("File references"),
    label_countable=ngettext("{count} file reference", "{count} file references"),
    public_facing=False,
)
class FileReference(Entity):
    """
    .. plugin:: entity:file-reference.
    """

    referee = BidirectionalToOne["FileReference", "HasFileReferences"](
        "betty.ancestry.has_file_references:HasFileReferences",
        "file_references",
        label=_("Referee"),
        description=_("The entity referencing the file"),
    )
    """
    The entity that references the file.
    """

    file = BidirectionalToOne["FileReference", File](
        "betty.ancestry.file:File",
        "referees",
        label=_("File"),
        description=_("The file being referenced"),
    )
    """
    The referenced file.
    """

    def __init__(
        self,
        referee: ToOneAssociate[HasFileReferences],
        file: ToOneAssociate[File],
        *,
        focus: FocusArea | None = None,
    ):
        super().__init__()
        self.referee = referee
        self.file = file
        self.focus = focus

    @property
    def focus(self) -> FocusArea | None:
        """
        The area within the 2-dimensional representation of the file to focus on.

        This can be used to locate where faces are in a photo, or a specific article in a newspaper scan, for example.
        """
        return self._focus

    @focus.setter
    def focus(self, focus: FocusArea | None) -> None:
        self._focus = focus



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING, Self, final, override

from betty.about import IS_DEVELOPMENT
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.locale import translation
from betty.locale.localizable.gettext import _
from betty.service.factory import Manufacturable
from betty.service.requirement import UnmetRequirement
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse
    from collections.abc import Iterable

    from betty.app import App
    from betty.plugin.discovery import ResolvableDiscovery


@final
@CommandDefinition(
    "dev-update-translations",
    label=_("Update all existing translations for Betty itself"),
)
class DevUpdateTranslations(Manufacturable, Command):
    """
    .. plugin:: command:dev-update-translations.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        return self._command_function

    async def _command_function(self) -> None:
        await translation.update_dev_translations()


def _discover(_) -> Iterable[ResolvableDiscovery[CommandDefinition]]:
    if not IS_DEVELOPMENT:
        raise UnmetRequirement("This is only available when developing Betty")
    yield DevUpdateTranslations



"""
Job execution.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from types import TracebackType


class Executor(ABC):
    """
    A job executor.
    """

    @abstractmethod
    async def start(self) -> None:
        """
        Start working on jobs.
        """

    @abstractmethod
    async def cancel(self) -> None:
        """
        Stop the executor and cancel any pending jobs.
        """

    @abstractmethod
    async def complete(self) -> None:
        """
        Wait for any pending jobs to complete and stop the executor.
        """

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if exc_val is None:
            await self.complete()
        else:
            await self.cancel()



"""
Samples are used to generate documentation about various parts of Betty.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import IntEnum, auto
from typing import TYPE_CHECKING, Generic, Self, TypeVar, final

from betty.locale.localizable import resolve_localizable

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator

    from ty_extensions import Intersection

    from betty.locale.localizable import Localizable, ResolvableLocalizable


@final
class Size(IntEnum):
    """
    A sample size indicator.
    """

    MINIMAL = auto()
    INTERMEDIATE = auto()
    FULL = auto()


@final
class Sample[T]:
    """
    A sample.

    Samples are useful for generating documentation and tests.
    """

    def __init__(
        self,
        subject: T,
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        size: Size = Size.INTERMEDIATE,
    ):
        self._subject = subject
        self._label = resolve_localizable(label)
        self._description = resolve_localizable(description) if description else None
        self._size = size

    @property
    def subject(self) -> T:
        """
        The sample subject.
        """
        return self._subject

    @property
    def label(self) -> Localizable:
        """
        The sample's human-readable short label.
        """
        return self._label

    @property
    def description(self) -> Localizable | None:
        """
        The sample's human-readable long description.
        """
        return self._description

    @property
    def size(self) -> Size:
        """
        The sample size.
        """
        return self._size


_SampleT = TypeVar("_SampleT", covariant=True)


@final
class Samples(
    Generic[_SampleT],  # noqa: UP046
):
    """
    A set of samples.
    """

    def __init__(
        self,
        samples: Iterable[
            Callable[[], Sample[_SampleT]]
            | Samples[_SampleT]
            | type[Intersection[_SampleT, Samplable]]
        ],
    ):
        self._samples = list(samples)

    def __iter__(self) -> Iterator[Sample[_SampleT]]:
        for sample in self._samples:
            if isinstance(sample, Samples):
                yield from sample
            elif isinstance(sample, type) and issubclass(sample, Samplable):
                yield from sample.samples()
            else:
                yield sample()

    def get(self, preferred_size: Size = Size.INTERMEDIATE, /):
        """
        Get a sample.
        """
        samples = sorted(self, key=lambda sample: abs(preferred_size - sample.size))
        if samples:
            return samples[0]
        raise SampleNotFound


class Samplable(ABC):
    """
    Allow a class to provide its own samples.
    """

    @classmethod
    @abstractmethod
    def samples(cls) -> Samples[Self]:
        """
        Get the samples.
        """


class SampleNotFound(Exception):
    """
    Raised when a sample could not be found.
    """



"""
Tools to build data types that have links.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from betty.locale.localizable.gettext import _
from betty.model import Entity
from betty.model.association import BidirectionalToManySingleType, ToManyAssociates

if TYPE_CHECKING:
    from betty.ancestry.link import Link


class HasLinks(Entity):
    """
    An entity that has associated :py:class:`betty.ancestry.link.Link` entities.
    """

    links = BidirectionalToManySingleType["HasLinks", "Link"](
        "betty.ancestry.link:Link",
        "owner",
        label=_("Links"),
        linked_data_embedded=True,
    )

    def __init__(
        self,
        *args: Any,
        links: ToManyAssociates[Link] | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if links is not None:
            self.links = links



"""
Key-value mapping record data types.
"""

from __future__ import annotations

from collections.abc import MutableMapping
from typing import Any, final

from betty.data.aggregate.record import RecordDefinition
from betty.data.indicator.selector import Key


@final
class TypedMappingDefinition[MutableMappingT: MutableMapping[str, Any]](
    RecordDefinition[MutableMappingT, Key]
):
    """
    A typed mapping definition.

    Actual values do not have to be :py:class:`typing.TypedDict`. They can be any mapping, but like typed dicts, values
    are limited to the defined elements.
    """



"""
Service plugin management.

Service levels can expose services of plugin instances.
"""

from __future__ import annotations

from importlib import metadata
from typing import TYPE_CHECKING, Any, cast, final, overload, override

from betty.collection.keyed import KeyedCollection
from betty.concurrent import AsynchronizedLock
from betty.machine_name import MachineName, ResolvableMachineName
from betty.plugin import (
    Plugin,
    PluginDefinition,
    ResolvableId,
    resolve_id,
)
from betty.plugin.discovery import ResolvableDiscovery
from betty.plugin.error import PluginNotFound
from betty.string import kebab_case_to_snake_case
from betty.typing import threadsafe

if TYPE_CHECKING:
    import builtins
    from collections.abc import AsyncIterator, Awaitable, Iterable, Iterator, Mapping

    from ty_extensions import Intersection

    from betty.service.level import ServiceLevel


@final
class PluginCollection[PluginDefinitionT: PluginDefinition, PluginT: Plugin](
    KeyedCollection[MachineName, ResolvableId[PluginDefinitionT], PluginT]
):
    """
    A collection of plugin instances.
    """

    def __init__(self, plugins: Iterable[Iterable[PluginT]], /):
        self._batches = tuple(map(tuple, plugins))
        self._all = {
            plugin.plugin().id: plugin for batch in self._batches for plugin in batch
        }

    @override
    def __len__(self) -> int:
        return len(self._all)

    @override
    def __iter__(self) -> Iterator[PluginT]:
        yield from self._all.values()

    @override
    def __contains__(self, key: Any) -> bool:
        if isinstance(key, type) and issubclass(key, Plugin):
            key = key.plugin().id
        elif isinstance(key, PluginDefinition):
            key = key.id
        return key in self._all

    @overload
    def __getitem__[T](
        self, key: type[Intersection[PluginT, T]]
    ) -> Intersection[PluginT, T]:
        pass

    @overload
    def __getitem__(self, key: ResolvableId[PluginDefinitionT]) -> PluginT:
        pass

    @override
    def __getitem__(self, key):
        return self._all[resolve_id(key)]

    @override
    def keys(self) -> Iterable[MachineName]:
        return self._all.keys()


@final
@threadsafe
class PluginManager[PluginDefinitionT: PluginDefinition]:
    """
    Expose the plugin type definition and plugin definitions for a specific plugin type.
    """

    def __init__(
        self,
        services: ServiceLevel,
        plugin_type: builtins.type[PluginDefinitionT],
        plugin_overrides: Iterable[ResolvableDiscovery[PluginDefinition]] | None = None,
        /,
    ):
        self._services = services
        self._type = plugin_type
        self._lock = AsynchronizedLock.new_threadsafe()
        self._discovery = (
            [self._discover] if plugin_overrides is None else plugin_overrides
        )
        self.__plugins: Mapping[MachineName, PluginDefinitionT] | None = None

    @property
    def type(self) -> builtins.type[PluginDefinitionT]:
        """
        The plugin type.
        """
        return self._type

    def _discover(
        self, services: ServiceLevel
    ) -> Iterable[ResolvableDiscovery[PluginDefinitionT]]:
        for entry_point in metadata.entry_points(
            group=f"betty.{kebab_case_to_snake_case(self.type.type().id)}"
        ):
            yield cast(ResolvableDiscovery[PluginDefinitionT], entry_point.load())

    async def _plugins(self) -> Mapping[MachineName, PluginDefinitionT]:
        from betty.plugin.discovery import discover

        if self.__plugins is not None:
            return self.__plugins
        async with self._lock:
            if self.__plugins is not None:
                return self.__plugins
            self.__plugins = {
                plugin.id: plugin
                for plugin in await discover(self._services, *self._discovery)
            }
            return self.__plugins

    async def __aiter__(self) -> AsyncIterator[PluginDefinitionT]:
        for plugin in (await self._plugins()).values():
            yield plugin

    async def _get(self, key: ResolvableMachineName) -> PluginDefinitionT:
        key = MachineName.resolve(key)
        try:
            return (await self._plugins())[key]
        except KeyError:
            raise PluginNotFound(self._type, key, await self.ids()) from None

    def __getitem__(self, key: ResolvableMachineName) -> Awaitable[PluginDefinitionT]:
        return self._get(key)

    async def ids(self) -> Iterable[MachineName]:
        """
        Iterate over the IDs of the available plugins.
        """
        return (await self._plugins()).keys()



"""
Data types describing persons.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override
from urllib.parse import quote

from betty.ancestry.has_citations import HasCitations
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_links import HasLinks
from betty.ancestry.has_notes import HasNotes
from betty.functools import unique
from betty.gender import GenderDefinition
from betty.gender.genders import Unknown as UnknownGender
from betty.json.linked_data import JsonLdObject, dump_context
from betty.locale.localizable.gettext import _, ngettext
from betty.model import EntityDefinition, persistent_id
from betty.model.association import BidirectionalToManySingleType, ToManyAssociates
from betty.model.schema import ToManySchema
from betty.plugin.schema import PluginIdSchema
from betty.privacy import HasPrivacy, Privacy

if TYPE_CHECKING:
    from collections.abc import Iterator, MutableSequence

    from betty.ancestry.citation import Citation
    from betty.ancestry.file_reference import FileReference
    from betty.ancestry.link import Link
    from betty.ancestry.note import Note
    from betty.ancestry.person_name import PersonName
    from betty.ancestry.presence import Presence
    from betty.gender import Gender
    from betty.locale.localizable import Localizable
    from betty.portable import PortableMapping
    from betty.project import Project


@final
@EntityDefinition(
    "person",
    label=_("Person"),
    label_plural=_("People"),
    label_countable=ngettext("{count} person", "{count} people"),
)
class Person(HasFileReferences, HasCitations, HasNotes, HasLinks, HasPrivacy):
    """
    .. plugin:: entity:person.
    """

    parents = BidirectionalToManySingleType["Person", "Person"](
        "betty.ancestry.person:Person",
        "children",
        label=_("Parents"),
    )
    """
    The person's parents.
    """

    children = BidirectionalToManySingleType["Person", "Person"](
        "betty.ancestry.person:Person",
        "parents",
        label=_("Children"),
    )
    """
    The person's children.
    """

    presences = BidirectionalToManySingleType["Person", "Presence"](
        "betty.ancestry.presence:Presence",
        "person",
        label=_("Presences"),
        description=_("This person's presences at events"),
        linked_data_embedded=True,
    )
    """
    The person's presences at events.
    """

    names = BidirectionalToManySingleType["Person", "PersonName"](
        "betty.ancestry.person_name:PersonName",
        "person",
        label=_("Names"),
        linked_data_embedded=True,
    )
    """
    The person's names.
    """

    def __init__(
        self,
        id: str | None = None,  # noqa: A002
        *,
        file_references: ToManyAssociates[FileReference] | None = None,
        citations: ToManyAssociates[Citation] | None = None,
        links: MutableSequence[Link] | None = None,
        notes: ToManyAssociates[Note] | None = None,
        privacy: Privacy | None = None,
        parents: ToManyAssociates[Person] | None = None,
        children: ToManyAssociates[Person] | None = None,
        presences: ToManyAssociates[Presence] | None = None,
        names: ToManyAssociates[PersonName] | None = None,
        gender: Gender | None = None,
    ):
        super().__init__(
            id,
            file_references=file_references,
            citations=citations,
            links=links,
            notes=notes,
            privacy=privacy,
        )
        if children is not None:
            self.children = children
        if parents is not None:
            self.parents = parents
        if presences is not None:
            self.presences = presences
        if names is not None:
            self.names = names
        self._gender = gender or UnknownGender()

    @property
    def gender(self) -> Gender:
        """
        The person's gender.
        """
        return self._gender

    @gender.setter
    def gender(self, gender: Gender) -> None:
        self._gender = gender

    @property
    def ancestors(self) -> Iterator[Person]:
        """
        All ancestors.
        """
        for parent in self.parents:
            yield parent
            yield from parent.ancestors

    @property
    def siblings(self) -> Iterator[Person]:
        """
        All siblings.
        """
        yield from unique(
            sibling
            for parent in self.parents
            for sibling in parent.children
            if sibling != self
        )

    @property
    def descendants(self) -> Iterator[Person]:
        """
        All descendants.
        """
        for child in self.children:
            yield child
            yield from child.descendants

    @override
    @property
    def label(self) -> Localizable:
        for name in self.names:
            if name.public:
                return name.label
        return super().label

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        url_generator = await project.url_generator
        dump_context(
            portable,
            names="https://schema.org/name",
            parents="https://schema.org/parent",
            children="https://schema.org/child",
            siblings="https://schema.org/sibling",
        )
        portable["@type"] = "https://schema.org/Person"
        portable["siblings"] = [
            url_generator.generate(
                f"betty-static:///person/{quote(sibling.id)}/index.json"
            )
            for sibling in self.siblings
            if persistent_id(sibling)
        ]
        if self.public:
            portable["gender"] = self.gender.plugin().id
        return portable

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "gender",
            PluginIdSchema(
                GenderDefinition.type(),
                [x async for x in project.plugins[GenderDefinition]],
            ),
            False,
        )
        schema.add_property("siblings", ToManySchema(title="Siblings"))
        return schema



from __future__ import annotations  # noqa: D100

import shutil
from asyncio import gather, to_thread
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Self, final, override

from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.locale.localizable.gettext import _
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse

    from betty.app import App

_LEGACY_CACHE_DIRECTORY_PATH = Path.home() / ".betty" / "cache"


@final
@CommandDefinition("clear-caches", label=_("Clear all caches"))
class ClearCaches(Manufacturable, Command):
    """
    .. plugin:: command:clear-caches.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        return self._command_function

    async def _command_function(self) -> None:
        await gather(self._app.cache.clear(), self._clear_legacy_cache())
        await self._app.user.message_information(_("All caches cleared."))

    async def _clear_legacy_cache(self) -> None:
        # Before Betty 0.5, Betty stored its caches in the home directory. Clear those until Betty 0.6.
        with suppress(FileNotFoundError):
            await to_thread(shutil.rmtree, _LEGACY_CACHE_DIRECTORY_PATH)



"""
Data types that have a locale.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, override

from betty.json.linked_data import (
    JsonLdObject,
    LinkedDataDumpableWithSchemaJsonLdObject,
)
from betty.json.schema import Null, OneOf
from betty.locale import HasLocale as StdHasLocale
from betty.locale import to_language_tag
from betty.locale.schema import LocaleSchema
from betty.privacy import is_public

if TYPE_CHECKING:
    from babel import Locale

    from betty.portable import PortableMapping
    from betty.project import Project


class HasLocale(StdHasLocale, LinkedDataDumpableWithSchemaJsonLdObject):
    """
    A resource that is localized, e.g. contains information in a specific locale.
    """

    @override
    @property
    def locale(self) -> Locale | None:
        return self._locale

    @locale.setter
    def locale(self, locale: Locale | None) -> None:
        self._locale = locale

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        portable["locale"] = to_language_tag(self.locale) if is_public(self) else None
        return portable

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property("locale", OneOf(LocaleSchema(), Null()))
        return schema



"""
Output progress to Rich.
"""

from typing import final, override

from rich.progress import Progress as _RichProgress

from betty.concurrent import AsynchronizedLock
from betty.progress import Progress
from betty.typing import threadsafe


@final
@threadsafe
class RichProgress(Progress):
    """
    Output progress to Rich.
    """

    def __init__(self, rich_progress: _RichProgress, rich_task_description: str):
        self._rich_progress = rich_progress
        self._rich_task = self._rich_progress.add_task(
            f"[green]{rich_task_description}", total=0
        )
        self._lock = AsynchronizedLock.new_threadsafe()
        self._total = 0

    @override
    async def add(self, add: int = 1, /) -> None:
        async with self._lock:
            self._total += add
            self._rich_progress.update(self._rich_task, total=self._total)

    @override
    async def done(self, done: int = 1, /) -> None:
        async with self._lock:
            self._rich_progress.update(self._rich_task, advance=done)



"""
Provide :py:class:`betty.copyright_notice.CopyrightNotice` plugins.
"""

from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, Self, final, override

import aiohttp

from betty.copyright_notice import CopyrightNotice, CopyrightNoticeDefinition
from betty.locale import DEFAULT_LOCALE, resolve_locale
from betty.locale.error import LocaleError
from betty.locale.localizable import resolve_localizable
from betty.locale.localizable.gettext import _
from betty.locale.localizable.plain import Plain
from betty.locale.localizable.static import StaticTranslations
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from betty.app import App
    from betty.locale.localizable import Localizable, ResolvableLocalizable
    from betty.project import Project


@final
@CopyrightNoticeDefinition("project-author", label=_("Project author"))
class ProjectAuthor(Manufacturable, CopyrightNotice):
    """
    .. plugin:: copyright-notice:project-author.
    """

    def __init__(self, author: ResolvableLocalizable | None):
        super().__init__()
        self._author = None if author is None else resolve_localizable(author)

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(project.configuration.author)

    @property
    @override
    def summary(self) -> Localizable:
        if self._author:
            return _("© Copyright {author}, unless otherwise credited").format(
                author=self._author
            )
        return _("© Copyright the author, unless otherwise credited")

    @property
    @override
    def text(self) -> Localizable:
        return self.summary


@final
@CopyrightNoticeDefinition("public-domain", label=_("Public domain"))
class PublicDomain(CopyrightNotice):
    """
    .. plugin:: copyright-notice:public-domain.
    """

    @property
    @override
    def summary(self) -> Localizable:
        return _("Public domain")

    @property
    @override
    def text(self) -> Localizable:
        return _(
            "Works in the public domain can be used or referenced without permission, because nobody holds any exclusive rights over these works (anymore)."
        )


@final
@CopyrightNoticeDefinition("streetmix", label=Plain("Streetmix LLC"))
class Streetmix(CopyrightNotice):
    """
    .. plugin:: copyright-notice:streetmix.
    """

    @override
    @property
    def summary(self) -> Localizable:
        return self.plugin().label

    @override
    @property
    def text(self) -> Localizable:
        return self.plugin().label

    @override
    @property
    def url(self) -> Localizable:
        return Plain("https://github.com/streetmix/streetmix")


def _copyright_url(language: str, page: str) -> str:
    return f"https://{language}.wikipedia.org/wiki/{page}"


@final
@CopyrightNoticeDefinition("wikipedia-contributors", label=_("Wikipedia contributors"))
class WikipediaContributors(Manufacturable, CopyrightNotice):
    """
    .. plugin:: copyright-notice:wikipedia-contributors.
    """

    def __init__(self, url: ResolvableLocalizable):
        super().__init__()
        self._url = resolve_localizable(url)

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        http_client = await app.http_client
        urls = {
            DEFAULT_LOCALE: _copyright_url("en", "Wikipedia:Copyrights"),
        }
        try:
            response = await http_client.get(
                "https://en.wikipedia.org/w/api.php?action=query&titles=Wikipedia:Copyrights&prop=langlinks&lllimit=500&format=json&formatversion=2"
            )
            response_json = await response.json()
        except aiohttp.ClientError:
            pass
        else:
            for link in response_json["query"]["pages"][0][
                "langlinks"
            ]:  # typing: ignore[index]
                # Wikipedia uses some languages that are not valid ISO codes, such as "simple".
                with suppress(LocaleError):
                    urls[resolve_locale(link["lang"])] = _copyright_url(
                        link["lang"], link["title"]
                    )
        return cls(StaticTranslations(urls))

    @override
    @property
    def summary(self) -> Localizable:
        return _("Copyright Wikipedia contributors")

    @override
    @property
    def text(self) -> Localizable:
        return _(
            "Copyright of these works lies with the original authors who contributed them to Wikipedia."
        )

    @override
    @property
    def url(self) -> Localizable:
        return self._url
